# -*- coding: utf-8 -*-
"""
    emmett.orm.helpers
    ------------------

    Provides ORM helpers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import copyreg
import operator
import re
import time

from functools import reduce, wraps
from typing import TYPE_CHECKING, Any, Callable

from pydal._globals import THREAD_LOCAL
from pydal.helpers.classes import ExecutionHandler
from pydal.objects import Field as _Field

from ..datastructures import sdict
from ..utils import cachedprop

if TYPE_CHECKING:
    from .objects import Table


class RowReferenceMeta:
    __slots__ = ['table', 'pk', 'caster']

    def __init__(self, table: Table, caster: Callable[[Any], Any]):
        self.table = table
        self.pk = table._id.name
        self.caster = caster

    def fetch(self, val):
        return self.table._db(self.table._id == self.caster(val)).select(
            limitby=(0, 1),
            orderby_on_limitby=False
        ).first()


class RowReferenceMultiMeta:
    __slots__ = ['table', 'pks', 'pks_idx', 'caster', 'casters']
    _casters = {'integer': int, 'string': str}

    def __init__(self, table: Table) -> None:
        self.table = table
        self.pks = list(table._primarykey)
        self.pks_idx = {key: idx for idx, key in enumerate(self.pks)}
        self.caster = tuple
        self.casters = {pk: self._casters[table[pk].type] for pk in self.pks}

    def fetch(self, val):
        query = reduce(
            operator.and_, [
                self.table[pk] == self.casters[pk](self.caster.__getitem__(val, idx))
                for pk, idx in self.pks_idx.items()
            ]
        )
        return self.table._db(query).select(
            limitby=(0, 1),
            orderby_on_limitby=False
        ).first()


class RowReferenceMixin:
    def _allocate_(self):
        if not self._refrecord:
            self._refrecord = self._refmeta.fetch(self)
        if not self._refrecord:
            raise RuntimeError(
                "Using a recursive select but encountered a broken " +
                "reference: %s %r" % (self._table, self)
            )

    def __getattr__(self, key: str) -> Any:
        if key == self._refmeta.pk:
            return self._refmeta.caster(self)
        if key in self._refmeta.table:
            self._allocate_()
        if self._refrecord:
            return self._refrecord.get(key, None)
        return None

    def get(self, key: str, default: Any = None) -> Any:
        return self.__getattr__(key, default)

    def __setattr__(self, key: str, value: Any):
        if key.startswith('_'):
            self._refmeta.caster.__setattr__(self, key, value)
            return
        self._allocate_()
        self._refrecord[key] = value

    def __getitem__(self, key):
        if key == self._refmeta.pk:
            return self._refmeta.caster(self)
        self._allocate_()
        return self._refrecord.get(key, None)

    def __setitem__(self, key, value):
        self._allocate_()
        self._refrecord[key] = value

    def __pure__(self):
        return self._refmeta.caster(self)

    def __repr__(self) -> str:
        return repr(self._refmeta.caster(self))


class RowReferenceInt(RowReferenceMixin, int):
    def __new__(cls, id, table: Table, *args: Any, **kwargs: Any):
        rv = super().__new__(cls, id, *args, **kwargs)
        int.__setattr__(rv, '_refmeta', RowReferenceMeta(table, int))
        int.__setattr__(rv, '_refrecord', None)
        return rv


class RowReferenceStr(RowReferenceMixin, str):
    def __new__(cls, id, table: Table, *args: Any, **kwargs: Any):
        rv = super().__new__(cls, id, *args, **kwargs)
        str.__setattr__(rv, '_refmeta', RowReferenceMeta(table, str))
        str.__setattr__(rv, '_refrecord', None)
        return rv


class RowReferenceMulti(RowReferenceMixin, tuple):
    def __new__(cls, id, table: Table, *args: Any, **kwargs: Any):
        tupid = tuple(id[key] for key in table._primarykey)
        rv = super().__new__(cls, tupid, *args, **kwargs)
        tuple.__setattr__(rv, '_refmeta', RowReferenceMultiMeta(table))
        tuple.__setattr__(rv, '_refrecord', None)
        return rv

    def __getattr__(self, key: str) -> Any:
        if key in self._refmeta.pks:
            return self._refmeta.casters[key](
                tuple.__getitem__(self, self._refmeta.pks_idx[key])
            )
        if key in self._refmeta.table:
            self._allocate_()
        if self._refrecord:
            return self._refrecord.get(key, None)
        return None

    def __getitem__(self, key):
        if key in self._refmeta.pks:
            return self._refmeta.casters[key](
                tuple.__getitem__(self, self._refmeta.pks_idx[key])
            )
        self._allocate_()
        return self._refrecord.get(key, None)


class GeoFieldWrapper(str):
    _rule_parens = re.compile(r"^(\(+)(?:.+)$")
    _json_geom_map = {
        "POINT": "Point",
        "LINESTRING": "LineString",
        "POLYGON": "Polygon",
        "MULTIPOINT": "MultiPoint",
        "MULTILINESTRING": "MultiLineString",
        "MULTIPOLYGON": "MultiPolygon"
    }

    def __new__(cls, value, *args: Any, **kwargs: Any):
        geometry, raw_coords = value.strip()[:-1].split("(", 1)
        rv = super().__new__(cls, value, *args, **kwargs)
        coords = cls._parse_coords_block(raw_coords)
        str.__setattr__(rv, '_geometry', geometry.strip())
        str.__setattr__(rv, '_coordinates', coords)
        return rv

    @classmethod
    def _parse_coords_block(cls, v):
        groups = []
        parens_match = cls._rule_parens.match(v)
        parens = parens_match.group(1) if parens_match else ''
        if parens:
            for element in v.split(parens):
                if not element:
                    continue
                element = element.strip()
                shift = -2 if element.endswith(",") else -1
                groups.append(f"{parens}{element}"[1:shift])
        if not groups:
            return cls._parse_coords_group(v)
        return tuple(
            cls._parse_coords_block(group) for group in groups
        )

    @staticmethod
    def _parse_coords_group(v):
        accum = []
        for element in v.split(","):
            accum.append(tuple(float(v) for v in element.split(" ")))
        return tuple(accum) if len(accum) > 1 else accum[0]

    def _repr_coords(self, val=None):
        val = val or self._coordinates
        if isinstance(val[0], tuple):
            accum = []
            for el in val:
                inner, plevel = self._repr_coords(el)
                inner = f"({inner})" if not plevel else inner
                accum.append(inner)
            return ",".join(accum), False
        return "%f %f" % val, True

    @property
    def geometry(self):
        return self._geometry

    @property
    def coordinates(self):
        return self._coordinates

    @property
    def groups(self):
        if not self._geometry.startswith("MULTI"):
            return tuple()
        return tuple(
            self.__class__(f"{self._geometry[5:]}({self._repr_coords(coords)[0]})")
            for coords in self._coordinates
        )

    def __json__(self):
        return {
            "type": self._json_geom_map[self._geometry],
            "coordinates": self._coordinates
        }


class PasswordFieldWrapper(str):
    _emt_field_hashed_contents_ = True


class Reference(object):
    def __init__(self, *args, **params):
        self.reference = [arg for arg in args]
        self.params = params
        self.refobj[id(self)] = self

    def __call__(self, func):
        if self.__class__.__name__ not in ['has_one', 'has_many']:
            raise SyntaxError(
                '%s cannot be used as a decorator' % self.__class__.__name__)
        if not callable(func):
            raise SyntaxError('Argument must be callable')
        if self.reference:
            raise SyntaxError(
                "When using %s as decorator, you must use the 'field' option" %
                self.__class__.__name__)
        new_reference = {func.__name__: {'method': func}}
        field = self.params.get('field')
        if field:
            new_reference[func.__name__]['field'] = field
        cast = self.params.get('cast')
        if cast:
            new_reference[func.__name__]['cast'] = cast
        self.reference = [new_reference]
        return self

    @property
    def refobj(self):
        return {}


class ReferenceData(sdict):
    def __init__(self, model_class, **kwargs):
        self.model_class = model_class
        super(ReferenceData, self).__init__(**kwargs)

    @cachedprop
    def dbset(self):
        if self.method:
            return self.method(self.model_class)
        return self.model_class.db

    @cachedprop
    def model_instance(self):
        if self.method:
            return self.dbset._model_._instance_()
        return self.dbset[self.model]._model_

    @property
    def table(self):
        return self.model_instance.table

    @property
    def table_name(self):
        return self.model_instance.tablename

    @property
    def fields_instances(self):
        return tuple(
            self.table[field]
            for field in self.model_instance._belongs_fks_.get(
                self.reverse, sdict(local_fields=[self.reverse])
            ).local_fields
        )


class RelationBuilder(object):
    def __init__(self, ref, model_instance):
        self.ref = ref
        self.model = model_instance

    def _make_refid(self, row):
        pks = self.model.primary_keys or ["id"]
        if row:
            return tuple(row[pk] for pk in pks)
        return tuple(self.model.table[pk] for pk in pks)

    def _extra_scopes(self, ref, model_instance=None):
        model_instance = model_instance or ref.model_instance
        rv = []
        if ref.scope is not None:
            scope_m = model_instance._scopes_[ref.scope].f
            rv.append(lambda f=scope_m, m=model_instance: f(m))
        if ref.where is not None:
            rv.append(lambda f=ref.where, m=model_instance: f(m))
        return rv

    def _patch_query_with_scopes(self, ref, query):
        for scope in self._extra_scopes(ref):
            query = query & scope()
        return query

    def _patch_query_with_scopes_on(self, ref, query, model_name):
        model = self.model.db[model_name]._model_
        for scope in self._extra_scopes(ref, model):
            query = query & scope()
        return query

    def _get_belongs(self, modelname, value):
        return self.model.db[modelname]._model_._belongs_fks_.get(value)

    def belongs_query(self):
        return reduce(
            operator.and_, [
                self.model.table[local] == self.model.db[self.ref.model][foreign]
                for local, foreign in self.ref.coupled_fields
            ]
        )

    @staticmethod
    def many_query(ref, rid):
        components = rid
        if ref.cast:
            components = []
            for element in rid:
                if isinstance(rid, _Field):
                    components.append(element.cast(ref.cast))
                else:
                    components.append(element)
        return reduce(
            operator.and_, [
                field == components[idx]
                for idx, field in enumerate(ref.fields_instances)
            ]
        )

    def _many(self, ref, rid):
        return ref.dbset.where(
            self._patch_query_with_scopes(ref, self.many_query(ref, rid))
        ).query

    def many(self, row=None):
        return self._many(self.ref, self._make_refid(row))

    def via(self, row=None):
        db = self.model.db
        rid = self._make_refid(row)
        sname = self.model.__class__.__name__
        stack = []
        midrel = self.model._hasmany_ref_.get(
            self.ref.via,
            self.model._hasone_ref_.get(self.ref.via)
        )
        stack.append(self.ref)
        while midrel.via is not None:
            stack.insert(0, midrel)
            midrel = self.model._hasmany_ref_.get(
                midrel.via,
                self.model._hasone_ref_.get(midrel.via)
            )
        query = self._many(midrel, rid)
        step_model = midrel.table_name
        sel_field = db[step_model].ALL
        last_belongs = None
        last_via = None
        for via in stack:
            rname = via.field or via.name[:-1]
            belongs_model = self._get_belongs(step_model, rname)
            if belongs_model:
                #: join table way
                last_belongs = step_model
                last_via = via
                _query = reduce(
                    operator.and_, [
                        (
                            db[belongs_model.model][foreign] ==
                            db[step_model][local]
                        ) for local, foreign in belongs_model.coupled_fields
                    ]
                )
                sel_field = db[belongs_model.model].ALL
                step_model = belongs_model.model
            else:
                #: shortcut way
                last_belongs = None
                rname = via.field or via.name
                midrel = db[step_model]._model_._hasmany_ref_.get(
                    rname,
                    db[step_model]._model_._hasone_ref_.get(rname)
                )
                if midrel.via:
                    nested = RelationBuilder(midrel, midrel.model_class)
                    nested_data = nested.via()
                    _query = nested_data[0]
                    step_model = midrel.model_class.tablename
                else:
                    _query = self._many(
                        midrel, [
                            db[step_model][step_field]
                            for step_field in (
                                db[step_model]._model_.primary_keys or ["id"]
                            )
                        ]
                    )
                    step_model = midrel.table_name
                sel_field = db[step_model].ALL
            query = query & _query
        query = via.dbset.where(
            self._patch_query_with_scopes_on(via, query, step_model)
        ).query
        return query, sel_field, sname, rid, last_belongs, last_via


class Callback(object):
    _inst_count_ = 0

    def __init__(self, f, t):
        self.t = []
        if isinstance(f, Callback):
            self.t += f.t
            f = f.f
        self.f = f
        self.t.append(t)
        self._inst_count_ = Callback._inst_count_
        Callback._inst_count_ += 1

    def __call__(self):
        return None


class TimingHandler(ExecutionHandler):
    def _timings(self):
        THREAD_LOCAL._emtdal_timings_ = getattr(
            THREAD_LOCAL, '_emtdal_timings_', [])
        return THREAD_LOCAL._emtdal_timings_

    @cachedprop
    def timings(self):
        return self._timings()

    def before_execute(self, command):
        self.t = time.time()

    def after_execute(self, command):
        dt = time.time() - self.t
        self.timings.append((command, dt))


class ConnectionContext:
    __slots__ = ['db', 'conn', 'with_transaction', 'reuse_if_open']

    def __init__(self, db, with_transaction=True, reuse_if_open=True):
        self.db = db
        self.conn = None
        self.with_transaction = with_transaction
        self.reuse_if_open = reuse_if_open

    def __enter__(self):
        self.conn = self.db.connection_open(
            with_transaction=self.with_transaction,
            reuse_if_open=self.reuse_if_open
        )
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            self.db.connection_close()
        self.conn = None

    async def __aenter__(self):
        self.conn = await self.db.connection_open_loop(
            with_transaction=self.with_transaction,
            reuse_if_open=self.reuse_if_open
        )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.conn:
            await self.db.connection_close_loop()
        self.conn = None


def decamelize(name):
    return "_".join(re.findall('[A-Z][^A-Z]*', name)).lower()


def camelize(name):
    return "".join(w.capitalize() for w in name.split("_"))


def make_tablename(classname):
    return decamelize(classname) + "s"


def wrap_scope_on_set(dbset, model_instance, scope):
    @wraps(scope)
    def wrapped(*args, **kwargs):
        return dbset.where(
            scope(model_instance, *args, **kwargs),
            model=model_instance.__class__)
    return wrapped


def wrap_scope_on_model(scope):
    @wraps(scope)
    def wrapped(cls, *args, **kwargs):
        return cls.db.where(
            scope(cls._instance_(), *args, **kwargs), model=cls)
    return wrapped


def wrap_virtual_on_model(model, virtual):
    @wraps(virtual)
    def wrapped(row, *args, **kwargs):
        return virtual(model, row, *args, **kwargs)
    return wrapped


def typed_row_reference(id: Any, table: Table):
    field_type = table._id.type if table._id else None
    return {
        'id': RowReferenceInt,
        'integer': RowReferenceInt,
        'string': RowReferenceStr,
        None: RowReferenceMulti
    }[field_type](id, table)


def typed_row_reference_from_record(record: Any, model: Any):
    field_type = model.table._id.type if model.table._id else None
    refcls = {
        'id': RowReferenceInt,
        'integer': RowReferenceInt,
        'string': RowReferenceStr,
        None: RowReferenceMulti
    }[field_type]
    if len(model._fieldset_pk) > 1:
        id = {pk: record[pk] for pk in model._fieldset_pk}
    else:
        id = record[tuple(model._fieldset_pk)[0]]
    rv = refcls(id, model.table)
    rv._refrecord = record
    return rv


def _rowref_pickler(obj):
    return obj._refmeta.caster, (obj.__pure__(), )


copyreg.pickle(RowReferenceInt, _rowref_pickler)
copyreg.pickle(RowReferenceStr, _rowref_pickler)
copyreg.pickle(RowReferenceMulti, _rowref_pickler)



# -*- coding: utf-8 -*-
"""
    tests.orm
    ---------

    Test pyDAL implementation over Emmett.
"""

import pytest

from datetime import datetime, timedelta
from uuid import uuid4

from pydal.objects import Table
from pydal import Field as _Field
from emmett import App, sdict, now
from emmett.orm import (
    Database, Field, Model,
    compute,
    before_insert, after_insert,
    before_update, after_update,
    before_delete, after_delete,
    before_save, after_save,
    before_destroy, after_destroy,
    before_commit, after_commit,
    rowattr, rowmethod,
    has_one, has_many, belongs_to, refers_to,
    scope
)
from emmett.orm.migrations.utils import generate_runtime_migration
from emmett.orm.objects import TransactionOps
from emmett.orm.errors import MissingFieldsForCompute
from emmett.validators import isntEmpty, hasLength


CALLBACK_OPS = {
    "before_insert": [],
    "before_update": [],
    "before_delete": [],
    "before_save":[],
    "before_destroy": [],
    "after_insert": [],
    "after_update": [],
    "after_delete": [],
    "after_save":[],
    "after_destroy": []
}
COMMIT_CALLBACKS = {
    "all": [],
    "insert": [],
    "update": [],
    "delete": [],
    "save": [],
    "destroy": []
}


def _represent_f(value):
    return value


def _widget_f(field, value):
    return value


class Stuff(Model):
    a = Field.string()
    b = Field()
    price = Field.float()
    quantity = Field.int()
    total = Field.float()
    total_watch = Field.float()
    invisible = Field()

    validation = {
        "a": {'presence': True},
        "total": {"allow": "empty"},
        "total_watch": {"allow": "empty"}
    }

    fields_rw = {
        "invisible": False
    }

    form_labels = {
        "a": "A label"
    }

    form_info = {
        "a": "A comment"
    }

    update_values = {
        "a": "a_update"
    }

    repr_values = {
        "a": _represent_f
    }

    form_widgets = {
        "a": _widget_f
    }

    @compute('total')
    def eval_total(self, row):
        return row.price * row.quantity

    @compute('total_watch', watch=['price', 'quantity'])
    def eval_total_watch(self, row):
        return row.price * row.quantity

    @before_insert
    def bi(self, fields):
        CALLBACK_OPS['before_insert'].append(fields)

    @after_insert
    def ai(self, fields, id):
        CALLBACK_OPS['after_insert'].append((fields, id))

    @before_update
    def bu(self, set, fields):
        CALLBACK_OPS['before_update'].append((set, fields))

    @after_update
    def au(self, set, fields):
        CALLBACK_OPS['after_update'].append((set, fields))

    @before_delete
    def bd(self, set):
        CALLBACK_OPS['before_delete'].append(set)

    @after_delete
    def ad(self, set):
        CALLBACK_OPS['after_delete'].append(set)

    @rowattr('totalv')
    def eval_total_v(self, row):
        return row.price * row.quantity

    @rowmethod('totalm')
    def eval_total_m(self, row):
        return row.price * row.quantity

    @classmethod
    def method_test(cls, t):
        return cls.db, cls.table, t


class Person(Model):
    has_many(
        'things', {'features': {'via': 'things'}}, {'pets': 'Dog.owner'},
        'subscriptions')

    name = Field()
    age = Field.int()


class Thing(Model):
    belongs_to('person')
    has_many('features')

    name = Field()
    color = Field()


class Feature(Model):
    belongs_to('thing')
    has_one('price')

    name = Field()


class Price(Model):
    belongs_to('feature')

    value = Field.int()


class Doctor(Model):
    has_many('appointments', {'patients': {'via': 'appointments'}})
    name = Field()


class Patient(Model):
    has_many('appointments', {'doctors': {'via': 'appointments'}})
    name = Field()


class Appointment(Model):
    belongs_to('patient', 'doctor')
    date = Field.datetime()


class User(Model):
    name = Field()
    has_many(
        'memberships', {'organizations': {'via': 'memberships'}},
        {'cover_orgs': {
            'via': 'memberships.organization',
            'where': lambda m: m.is_cover == True}})


class Organization(Model):
    name = Field()
    is_cover = Field.bool(default=False)

    @has_many()
    def admin_memberships3(self):
        return Membership.admins()

    has_many(
        'memberships', {'users': {'via': 'memberships'}},
        {'admin_memberships': {'target': 'Membership', 'scope': 'admins'}},
        {'admins': {'via': 'admin_memberships.user'}},
        {'admin_memberships2': {
            'target': 'Membership', 'where': lambda m: m.role == 'admin'}},
        {'admins2': {'via': 'admin_memberships2.user'}},
        {'admins3': {'via': 'admin_memberships3.user'}})


class Membership(Model):
    belongs_to('user', 'organization')
    role = Field()

    @scope('admins')
    def filter_admins(self):
        return self.role == 'admin'


class House(Model):
    name = Field()


class Mouse(Model):
    tablename = "mice"
    has_many('elephants')
    name = Field()


class NeedSplit(Model):
    name = Field()


class Zoo(Model):
    has_many('animals', 'elephants', {'mice': {'via': 'elephants.mouse'}})
    name = Field()


class Animal(Model):
    belongs_to('zoo')
    name = Field()

    @rowattr('doublename')
    def get_double_name(self, row):
        return row.name * 2

    @rowattr('pretty')
    def get_pretty(self, row):
        return row.name

    @before_insert
    def bi(self, *args, **kwargs):
        pass

    @before_insert
    def bi2(self, *args, **kwargs):
        pass


class Elephant(Animal):
    belongs_to('mouse')
    color = Field()

    @rowattr('pretty')
    def get_pretty(self, row):
        return row.name + " " + row.color

    @before_insert
    def bi2(self, *args, **kwargs):
        pass


class Dog(Model):
    belongs_to({'owner': 'Person'})
    name = Field()


class Subscription(Model):
    belongs_to('person')

    name = Field()
    status = Field.int()
    expires_at = Field.datetime()

    STATUS = {'active': 1, 'suspended': 2, 'other': 3}

    @scope('expired')
    def get_expired(self):
        return self.expires_at < datetime.now()

    @scope('of_status')
    def filter_status(self, *statuses):
        if len(statuses) == 1:
            return self.status == self.STATUS[statuses[0]]
        return self.status.belongs(*[self.STATUS[v] for v in statuses])


class Product(Model):
    name = Field.string()
    price = Field.float(default=0.0)


class Cart(Model):
    has_many({"elements": "CartElement"})

    updated_at = Field.datetime(default=now, update=now)
    total_denorm = Field.float(default=0.0)
    revision = Field.string(default=lambda: uuid4().hex, update=lambda: uuid4().hex)

    def _sum_elements(self, row):
        summable = (CartElement.quantity.cast("float") * Product.price).sum()
        sum = row.elements.join("product").select(summable).first()
        return sum[summable] or 0.0

    @before_save
    def _rebuild_total(self, row):
        row.total_denorm = self._sum_elements(row)

    @rowattr("total")
    def _compute_total(self, row):
        return self._sum_elements(row)


class CartElement(Model):
    belongs_to("product", "cart")

    updated_at = Field.datetime(default=now, update=now)
    quantity = Field.int(default=1)
    price_denorm = Field.float(default=0.0)

    @before_save
    def _rebuild_price(self, row):
        row.price_denorm = row.quantity * row.product.price

    @after_save
    def _refresh_cart_after_update(self, row):
        row.cart.save()

    @before_destroy
    def _undo_quantity_on_removal(self, row):
        row.quantity = 0
        row.price_denorm = 0

    @after_destroy
    def _refresh_cart_after_removal(self, row):
        row.cart.save()

    @rowattr("price")
    def _compute_price(self, row):
        return row.quantity * row.product.price


class SelfRef(Model):
    refers_to({'parent': 'self'})

    name = Field.string()


class CustomPKType(Model):
    id = Field.string()


class CustomPKName(Model):
    primary_keys = ["name"]
    name = Field.string()


class CustomPKMulti(Model):
    primary_keys = ["first_name", "last_name"]
    first_name = Field.string()
    last_name = Field.string()


class CommitWatcher(Model):
    foo = Field.string()
    created_at = Field.datetime(default=now)
    updated_at = Field.datetime(default=now, update=now)

    @before_commit
    def _commit_watch_before(self, op_type, ctx):
        COMMIT_CALLBACKS["all"].append(("before", op_type, ctx))

    @after_commit
    def _commit_watch_after(self, op_type, ctx):
        COMMIT_CALLBACKS["all"].append(("after", op_type, ctx))

    @before_commit.operation(TransactionOps.save)
    def _commit_watch_before_save(self, ctx):
        COMMIT_CALLBACKS["save"].append(("before", ctx))

    @after_commit.operation(TransactionOps.save)
    def _commit_watch_after_save(self, ctx):
        COMMIT_CALLBACKS["save"].append(("after", ctx))

    @before_commit.operation(TransactionOps.destroy)
    def _commit_watch_before_destroy(self, ctx):
        COMMIT_CALLBACKS["destroy"].append(("before", ctx))

    @after_commit.operation(TransactionOps.destroy)
    def _commit_watch_after_destroy(self, ctx):
        COMMIT_CALLBACKS["destroy"].append(("after", ctx))


@pytest.fixture(scope='module')
def _db():
    app = App(__name__)
    db = Database(
        app, config=sdict(
            uri=f'sqlite://{uuid4().hex}.db',
            auto_connect=True
        )
    )
    db.define_models(
        Stuff, Person, Thing, Feature, Price, Dog, Subscription,
        Doctor, Patient, Appointment,
        User, Organization, Membership,
        House, Mouse, NeedSplit, Zoo, Animal, Elephant,
        Product, Cart, CartElement,
        SelfRef,
        CustomPKType, CustomPKName, CustomPKMulti,
        CommitWatcher
    )
    return db


@pytest.fixture(scope='function')
def db(_db):
    migration = generate_runtime_migration(_db)
    migration.up()
    yield _db
    migration.down()


def test_db_instance(db):
    assert isinstance(db, Database)


def test_table_definition(db):
    assert isinstance(db.Stuff, Table)
    assert isinstance(db[Stuff.tablename], Table)


def test_fields(db):
    assert isinstance(db.Stuff.a, _Field)
    assert db.Stuff.a.type == "string"


def test_validators(db):
    assert len(db.Stuff.a.requires) == 2
    assert isinstance(db.Stuff.a.requires[0], isntEmpty)
    assert isinstance(db.Stuff.a.requires[1], hasLength)


def test_visibility(db):
    assert db.Stuff.a.readable is True
    assert db.Stuff.a.writable is True
    assert db.Stuff.invisible.readable is False
    assert db.Stuff.invisible.writable is False


def test_labels(db):
    assert db.Stuff.a.label == "A label"


def test_comments(db):
    assert db.Stuff.a.comment == "A comment"


def test_updates(db):
    assert db.Stuff.a.update == "a_update"


def test_representation(db):
    assert db.Stuff.a.represent == _represent_f


def test_widgets(db):
    assert db.Stuff.a.widget == _widget_f


# def test_set_helper(db):
#     assert isinstance(db.Stuff.b.requires, notInDb)


def test_computations(db):
    #: no watch
    row = sdict(price=12.95, quantity=3)
    rv = db.Stuff.total.compute(row)
    assert rv == 12.95 * 3
    #: watch fulfill
    row = sdict(price=12.95, quantity=3)
    rv = db.Stuff.total_watch.compute(row)
    assert rv == 12.95 * 3
    #: watch missing field
    row = sdict(price=12.95)
    with pytest.raises(MissingFieldsForCompute):
        db.Stuff.total_watch.compute(row)
    #: update flow
    res = Stuff.create(a="foo", price=12.95, quantity=1)
    row = Stuff.get(res.id)
    with pytest.raises(MissingFieldsForCompute):
        row.update_record(quantity=2)
    row.update_record(price=row.price, quantity=2)
    assert row.total == row.price * 2
    assert row.total_watch == row.price * 2


def test_callbacks(db):
    fields = {"a": 1, "b": 2, "c": 3}
    id = 12
    db.Stuff._before_insert[-1](fields)
    assert CALLBACK_OPS["before_insert"][-1] == fields
    db.Stuff._after_insert[-1](fields, id)
    res = CALLBACK_OPS["after_insert"][-1]
    assert res[0] == fields and res[1] == id
    set = {"a": "b"}
    db.Stuff._before_update[-1](set, fields)
    res = CALLBACK_OPS["before_update"][-1]
    assert res[0] == set and res[1] == fields
    db.Stuff._after_update[-1](set, fields)
    res = CALLBACK_OPS["after_update"][-1]
    assert res[0] == set and res[1] == fields
    db.Stuff._before_delete[-1](set)
    res = CALLBACK_OPS["before_delete"][-1]
    assert res == set
    db.Stuff._after_delete[-1](set)
    res = CALLBACK_OPS["after_delete"][-1]
    assert res == set


def test_save(db):
    p1 = db.Product.insert(name="foo", price=2.99)
    p2 = db.Product.insert(name="bar", price=7.49)
    cart = db.Cart.insert()
    assert cart.total == 0
    assert cart.total_denorm == 0

    cart_rev = cart.revision
    item = CartElement.new(cart=cart, product=p1)
    item.save()
    assert item.price == p1.price
    assert item.price_denorm == p1.price
    cart = Cart.get(cart.id)
    assert cart.total == p1.price
    assert cart.total_denorm == p1.price
    assert cart.revision != cart_rev

    cart_rev = cart.revision
    item = CartElement.new(cart=cart, product=p2, quantity=3)
    item.save()
    assert item.price == p2.price * 3
    assert item.price_denorm == p2.price * 3
    cart = Cart.get(cart.id)
    assert cart.total == p1.price + p2.price * 3
    assert cart.total_denorm == p1.price + p2.price * 3
    assert cart.revision != cart_rev


def test_destroy(db):
    p1 = db.Product.insert(name="foo", price=2.99)
    p2 = db.Product.insert(name="bar", price=7.49)
    cart = db.Cart.insert()

    item = CartElement.new(cart=cart, product=p1)
    item.save()
    item = CartElement.new(cart=cart, product=p2, quantity=3)
    item.save()

    cart = Cart.get(cart.id)
    cart_rev = cart.revision

    item.destroy()
    assert not item.id
    assert not item.price
    assert not item.price_denorm
    cart = Cart.get(cart.id)
    assert cart.total == p1.price
    assert cart.total_denorm == p1.price
    assert cart.revision != cart_rev


def test_commit_callbacks(db):
    #: insert
    row = db.CommitWatcher.insert(foo="test1")
    assert not COMMIT_CALLBACKS["all"]
    db.commit()

    assert len(COMMIT_CALLBACKS["all"]) == 2

    before, after = COMMIT_CALLBACKS["all"]

    order, op_type, ctx = before
    assert order == "before"
    assert op_type == TransactionOps.insert
    assert ctx.values.foo == "test1"
    assert ctx.return_value == row.id

    order, op_type, ctx = after
    assert order == "after"
    assert op_type == TransactionOps.insert
    assert ctx.values.foo == "test1"
    assert ctx.return_value == row.id

    COMMIT_CALLBACKS["all"].clear()

    #: update
    row.update_record(foo="test1a")
    assert not COMMIT_CALLBACKS["all"]
    db.commit()

    assert len(COMMIT_CALLBACKS["all"]) == 2

    before, after = COMMIT_CALLBACKS["all"]

    order, op_type, ctx = before
    assert order == "before"
    assert op_type == TransactionOps.update
    assert ctx.dbset
    assert ctx.values.foo == "test1a"
    assert ctx.return_value == 1

    order, op_type, ctx = after
    assert order == "after"
    assert op_type == TransactionOps.update
    assert ctx.dbset
    assert ctx.values.foo == "test1a"
    assert ctx.return_value == 1

    COMMIT_CALLBACKS["all"].clear()

    #: delete
    row.delete_record()
    assert not COMMIT_CALLBACKS["all"]
    db.commit()

    assert len(COMMIT_CALLBACKS["all"]) == 2

    before, after = COMMIT_CALLBACKS["all"]

    order, op_type, ctx = before
    assert order == "before"
    assert op_type == TransactionOps.delete
    assert ctx.dbset
    assert ctx.return_value == 1

    order, op_type, ctx = after
    assert order == "after"
    assert op_type == TransactionOps.delete
    assert ctx.dbset
    assert ctx.return_value == 1

    COMMIT_CALLBACKS["all"].clear()

    #: save:insert
    row = CommitWatcher.new(foo="test2")
    row.save()
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["save"]
    db.commit()

    assert len(COMMIT_CALLBACKS["all"]) == 4
    assert len(COMMIT_CALLBACKS["save"]) == 2

    before_ins, before_save, after_ins, after_save = COMMIT_CALLBACKS["all"]

    order, op_type, ctx = before_ins
    assert order == "before"
    assert op_type == TransactionOps.insert
    assert ctx.values.foo == "test2"
    assert ctx.return_value == row.id

    order, op_type, ctx = after_ins
    assert order == "after"
    assert op_type == TransactionOps.insert
    assert ctx.values.foo == "test2"
    assert ctx.return_value == row.id

    order, op_type, ctx = before_save
    assert order == "before"
    assert op_type == TransactionOps.save
    assert ctx.values.foo == "test2"
    assert ctx.return_value == row.id
    assert ctx.row.id == row.id
    assert "id" in ctx.changes

    order, op_type, ctx = after_save
    assert order == "after"
    assert op_type == TransactionOps.save
    assert ctx.values.foo == "test2"
    assert ctx.return_value == row.id
    assert ctx.row.id == row.id
    assert "id" in ctx.changes

    before_save, after_save = COMMIT_CALLBACKS["save"]

    order, ctx = before_save
    assert order == "before"
    assert ctx.values.foo == "test2"
    assert ctx.return_value == row.id
    assert ctx.row.id == row.id
    assert "id" in ctx.changes

    order, ctx = after_save
    assert order == "after"
    assert ctx.values.foo == "test2"
    assert ctx.return_value == row.id
    assert ctx.row.id == row.id
    assert "id" in ctx.changes

    COMMIT_CALLBACKS["all"].clear()
    COMMIT_CALLBACKS["save"].clear()

    #: save:update
    row.foo = "test2a"
    row.save()
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["save"]
    db.commit()

    assert len(COMMIT_CALLBACKS["all"]) == 4
    assert len(COMMIT_CALLBACKS["save"]) == 2

    before_upd, before_save, after_upd, after_save = COMMIT_CALLBACKS["all"]

    order, op_type, ctx = before_upd
    assert order == "before"
    assert op_type == TransactionOps.update
    assert ctx.dbset
    assert ctx.values.foo == "test2a"
    assert ctx.return_value == 1

    order, op_type, ctx = after_upd
    assert order == "after"
    assert op_type == TransactionOps.update
    assert ctx.dbset
    assert ctx.values.foo == "test2a"
    assert ctx.return_value == 1

    order, op_type, ctx = before_save
    assert order == "before"
    assert op_type == TransactionOps.save
    assert ctx.dbset
    assert ctx.values.foo == "test2a"
    assert ctx.return_value == 1
    assert ctx.row.id == row.id
    assert set(ctx.changes.keys()).issubset({"foo", "updated_at"})

    order, op_type, ctx = after_save
    assert order == "after"
    assert op_type == TransactionOps.save
    assert ctx.dbset
    assert ctx.values.foo == "test2a"
    assert ctx.return_value == 1
    assert ctx.row.id == row.id
    assert set(ctx.changes.keys()).issubset({"foo", "updated_at"})

    before_save, after_save = COMMIT_CALLBACKS["save"]

    order, ctx = before_save
    assert order == "before"
    assert ctx.dbset
    assert ctx.values.foo == "test2a"
    assert ctx.return_value == 1
    assert ctx.row.id == row.id
    assert set(ctx.changes.keys()).issubset({"foo", "updated_at"})

    order, ctx = after_save
    assert order == "after"
    assert ctx.dbset
    assert ctx.values.foo == "test2a"
    assert ctx.return_value == 1
    assert ctx.row.id == row.id
    assert set(ctx.changes.keys()).issubset({"foo", "updated_at"})

    COMMIT_CALLBACKS["all"].clear()
    COMMIT_CALLBACKS["save"].clear()

    #: destroy
    row.destroy()
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["destroy"]
    db.commit()

    assert len(COMMIT_CALLBACKS["all"]) == 4
    assert len(COMMIT_CALLBACKS["destroy"]) == 2

    before_del, before_destroy, after_del, after_destroy = COMMIT_CALLBACKS["all"]

    order, op_type, ctx = before_del
    assert order == "before"
    assert op_type == TransactionOps.delete
    assert ctx.dbset
    assert ctx.return_value == 1

    order, op_type, ctx = after_del
    assert order == "after"
    assert op_type == TransactionOps.delete
    assert ctx.dbset
    assert ctx.return_value == 1

    order, op_type, ctx = before_destroy
    assert order == "before"
    assert op_type == TransactionOps.destroy
    assert ctx.dbset
    assert ctx.return_value == 1
    assert ctx.row.id == row.id

    order, op_type, ctx = after_destroy
    assert order == "after"
    assert op_type == TransactionOps.destroy
    assert ctx.dbset
    assert ctx.return_value == 1
    assert ctx.row.id == row.id

    before_destroy, after_destroy = COMMIT_CALLBACKS["destroy"]

    order, ctx = before_destroy
    assert order == "before"
    assert ctx.dbset
    assert ctx.return_value == 1
    assert ctx.row.id == row.id

    order, ctx = after_destroy
    assert order == "after"
    assert ctx.dbset
    assert ctx.return_value == 1
    assert ctx.row.id == row.id

    COMMIT_CALLBACKS["all"].clear()
    COMMIT_CALLBACKS["destroy"].clear()


def test_callbacks_skip(db):
    for stack in CALLBACK_OPS.values():
        stack.clear()
    for stack in COMMIT_CALLBACKS.values():
        stack.clear()

    #: insert
    row = db.CommitWatcher.insert(foo="test1", skip_callbacks=True)
    db.commit()
    assert not CALLBACK_OPS["before_insert"]
    assert not CALLBACK_OPS["after_insert"]
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["insert"]

    #: update
    row.update_record(foo="test1a", skip_callbacks=True)
    db.commit()
    assert not CALLBACK_OPS["before_update"]
    assert not CALLBACK_OPS["after_update"]
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["update"]

    #: delete
    row.delete_record(skip_callbacks=True)
    db.commit()
    assert not CALLBACK_OPS["before_delete"]
    assert not CALLBACK_OPS["after_delete"]
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["delete"]

    #: save:insert
    row = CommitWatcher.new(foo="test2")
    row.save(skip_callbacks=True)
    db.commit()
    assert not CALLBACK_OPS["before_save"]
    assert not CALLBACK_OPS["after_save"]
    assert not CALLBACK_OPS["before_insert"]
    assert not CALLBACK_OPS["after_insert"]
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["save"]
    assert not COMMIT_CALLBACKS["insert"]

    #: save:update
    row.foo = "test2a"
    row.save(skip_callbacks=True)
    db.commit()
    assert not CALLBACK_OPS["before_save"]
    assert not CALLBACK_OPS["after_save"]
    assert not CALLBACK_OPS["before_update"]
    assert not CALLBACK_OPS["after_update"]
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["save"]
    assert not COMMIT_CALLBACKS["update"]

    #: destroy
    row.destroy(skip_callbacks=True)
    db.commit()
    assert not CALLBACK_OPS["before_destroy"]
    assert not CALLBACK_OPS["after_destroy"]
    assert not CALLBACK_OPS["before_delete"]
    assert not CALLBACK_OPS["after_delete"]
    assert not COMMIT_CALLBACKS["all"]
    assert not COMMIT_CALLBACKS["destroy"]
    assert not COMMIT_CALLBACKS["delete"]


def test_rowattrs(db):
    db.Stuff._before_insert = []
    db.Stuff._after_insert = []
    res = db.Stuff.insert(a="foo", b="bar", price=12.95, quantity=3)
    db.commit()
    row = Stuff.get(res)
    assert row.totalv == 12.95 * 3


def test_rowmethods(db):
    db.Stuff._before_insert = []
    db.Stuff._after_insert = []
    res = db.Stuff.insert(a="foo", b="bar", price=12.95, quantity=3)
    db.commit()
    row = Stuff.get(res)
    assert row.totalm() == 12.95 * 3


def test_modelmethods(db):
    tm = "foo"
    rv = Stuff.method_test(tm)
    assert rv[0] == db and rv[1] == db.Stuff and rv[2] == tm


def test_relations(db):
    p1 = db.Person.insert(name="Giovanni", age=25)
    p2 = db.Person.insert(name="Giorgio", age=30)
    t1 = db.Thing.insert(name="apple", color="red", person=p1)
    t2 = db.Thing.insert(name="apple", color="green", person=p1)
    f = db.Feature.insert(name="tasty", thing=t1)
    db.Price.insert(value=5, feature=f)
    #: belongs, has_one, has_many
    p1 = db.Person(name="Giovanni")
    p2 = db.Person(name="Giorgio")
    assert p1.things.count() == 2
    assert p2.things.count() == 0
    t = p1.things.where(lambda t: t.color == "red").select()
    assert len(t) == 1
    assert t[0].name == "apple" and t[0].color == "red" and t[0].person.id == p1.id
    f = p1.things()[0].features()
    assert len(f) == 1
    assert f[0].name == "tasty" and f[0].thing.id == t[0].id and \
        f[0].thing.person.id == p1.id
    m = p1.things()[0].features()[0].price()
    assert (
        m.value == 5 and m.feature.id == f[0].id and
        m.feature.thing.id == t[0].id and m.feature.thing.person.id == p1.id
    )
    p2.things.add(t2)
    assert p1.things.count() == 1
    assert p2.things.count() == 1
    t = p2.things()
    assert len(t) == 1
    assert t[0].name == "apple" and t[0].color == "green" and t[0].person.id == p2.id
    p2.things.remove(t2)
    assert p2.things.count() == 0
    assert db(db.things).count() == 1
    #: has_many via as shortcut
    assert len(p1.features()) == 1
    #: has_many via with join tables logic
    doctor = db.Doctor.insert(name="cox")
    patient = db.Patient.insert(name="mario")
    db.Appointment.insert(doctor=1, patient=1)
    assert len(doctor.patients()) == 1
    assert len(doctor.appointments()) == 1
    assert len(patient.doctors()) == 1
    assert len(patient.appointments()) == 1
    joe = db.User.insert(name='joe')
    jim = db.User.insert(name='jim')
    org = db.Organization.insert(name='')
    org.users.add(joe, role='admin')
    org.users.add(jim, role='manager')
    assert len(org.users()) == 2
    assert len(joe.organizations()) == 1
    assert len(jim.organizations()) == 1
    assert joe.organizations().first().id == org
    assert jim.organizations().first().id == org
    assert joe.memberships().first().role == 'admin'
    assert jim.memberships().first().role == 'manager'
    org.users.remove(joe)
    org.users.remove(jim)
    assert len(org.users(reload=True)) == 0
    assert len(joe.organizations(reload=True)) == 0
    assert len(jim.organizations(reload=True)) == 0
    #: has_many with specified feld
    db.Dog.insert(name='pongo', owner=p1)
    assert len(p1.pets()) == 1 and p1.pets().first().name == 'pongo'
    #: has_many via with specified field
    zoo = db.Zoo.insert(name='magic zoo')
    mouse = db.Mouse.insert(name='jerry')
    db.Elephant.insert(name='dumbo', color='pink', mouse=mouse, zoo=zoo)
    assert len(zoo.mice()) == 1


def test_tablenames(db):
    assert db.House == db.houses
    assert db.Mouse == db.mice
    assert db.NeedSplit == db.need_splits


def test_inheritance(db):
    assert 'name' in db.Animal.fields
    assert 'name' in db.Elephant.fields
    assert 'zoo' in db.Animal.fields
    assert 'zoo' in db.Elephant.fields
    assert 'color' in db.Elephant.fields
    assert 'color' not in db.Animal.fields
    assert Elephant._all_virtuals_['get_double_name'] is \
        Animal._all_virtuals_['get_double_name']
    assert Elephant._all_virtuals_['get_pretty'] is not \
        Animal._all_virtuals_['get_pretty']
    assert Elephant._all_callbacks_['bi'] is \
        Animal._all_callbacks_['bi']
    assert Elephant._all_callbacks_['bi2'] is not \
        Animal._all_callbacks_['bi2']


def test_scopes(db):
    p = db.Person.insert(name="Walter", age=50)
    s = db.Subscription.insert(
        name="a", expires_at=datetime.now() - timedelta(hours=20), person=p,
        status=1)
    s2 = db.Subscription.insert(
        name="b", expires_at=datetime.now() + timedelta(hours=20), person=p,
        status=2)
    db.Subscription.insert(
        name="c", expires_at=datetime.now() + timedelta(hours=20), person=p,
        status=3)
    rows = db(db.Subscription).expired().select()
    assert len(rows) == 1 and rows[0].id == s
    rows = p.subscriptions.expired().select()
    assert len(rows) == 1 and rows[0].id == s
    rows = Subscription.expired().select()
    assert len(rows) == 1 and rows[0].id == s
    rows = db(db.Subscription).of_status('active', 'suspended').select()
    assert len(rows) == 2 and rows[0].id == s and rows[1].id == s2
    rows = p.subscriptions.of_status('active', 'suspended').select()
    assert len(rows) == 2 and rows[0].id == s and rows[1].id == s2
    rows = Subscription.of_status('active', 'suspended').select()
    assert len(rows) == 2 and rows[0].id == s and rows[1].id == s2


def test_relations_scopes(db):
    gus = db.User.insert(name="Gus Fring")
    org = db.Organization.insert(name="Los pollos hermanos")
    org.users.add(gus, role="admin")
    frank = db.User.insert(name="Frank")
    org.users.add(frank, role='manager')
    assert org.admins.count() == 1
    assert org.admins2.count() == 1
    assert org.admins3.count() == 1
    org.users.remove(gus)
    org.users.remove(frank)
    assert org.admins.count() == 0
    assert org.admins2.count() == 0
    assert org.admins3.count() == 0
    org2 = db.Organization.insert(name="Laundry", is_cover=True)
    org2.users.add(gus, role="admin")
    assert gus.cover_orgs.count() == 1
    assert gus.cover_orgs().first().id == org2
    org2.users.remove(gus)
    assert gus.cover_orgs.count() == 0
    org.delete_record()
    org2.delete_record()
    #: creation/addition
    org = db.Organization.insert(name="Los pollos hermanos")
    org.admins.add(gus)
    assert org.admins.count() == 1
    org.admins.remove(gus)
    assert org.admins.count() == 0
    org.delete_record()
    org = db.Organization.insert(name="Los pollos hermanos")
    org.admins2.add(gus)
    assert org.admins2.count() == 1
    org.admins2.remove(gus)
    assert org.admins2.count() == 0
    org.delete_record()
    org = db.Organization.insert(name="Los pollos hermanos")
    org.admins3.add(gus)
    assert org.admins3.count() == 1
    org.admins3.remove(gus)
    assert org.admins3.count() == 0
    org.delete_record()
    gus = User.get(name="Gus Fring")
    org2 = db.Organization.insert(name="Laundry", is_cover=True)
    gus.cover_orgs.add(org2)
    assert gus.cover_orgs.count() == 1
    assert gus.cover_orgs().first().id == org2
    gus.cover_orgs.remove(org2)
    assert gus.cover_orgs.count() == 0


def test_model_where(db):
    assert Subscription.where(lambda s: s.status == 1).query == \
        db(db.Subscription.status == 1).query


def test_model_first(db):
    p = db.Person.insert(name="Walter", age=50)
    db.Subscription.insert(
        name="a",
        expires_at=datetime.now() + timedelta(hours=20),
        person=p,
        status=1
    )
    db.Subscription.insert(
        name="b",
        expires_at=datetime.now() + timedelta(hours=20),
        person=p,
        status=1
    )
    db.CustomPKType.insert(id="a")
    db.CustomPKType.insert(id="b")
    db.CustomPKName.insert(name="a")
    db.CustomPKName.insert(name="b")
    db.CustomPKMulti.insert(first_name="foo", last_name="bar")
    db.CustomPKMulti.insert(first_name="foo", last_name="baz")
    db.CustomPKMulti.insert(first_name="bar", last_name="baz")

    assert Subscription.first().id == Subscription.all().select(
        orderby=Subscription.id,
        limitby=(0, 1)
    ).first().id
    assert CustomPKType.first().id == CustomPKType.all().select(
        orderby=CustomPKType.id,
        limitby=(0, 1)
    ).first().id
    assert CustomPKName.first().name == CustomPKName.all().select(
        orderby=CustomPKName.name,
        limitby=(0, 1)
    ).first().name
    assert CustomPKMulti.first() == CustomPKMulti.all().select(
        orderby=CustomPKMulti.first_name|CustomPKMulti.last_name,
        limitby=(0, 1)
    ).first()


def test_model_last(db):
    p = db.Person.insert(name="Walter", age=50)
    db.Subscription.insert(
        name="a",
        expires_at=datetime.now() + timedelta(hours=20),
        person=p,
        status=1
    )
    db.Subscription.insert(
        name="b",
        expires_at=datetime.now() + timedelta(hours=20),
        person=p,
        status=1
    )
    db.CustomPKType.insert(id="a")
    db.CustomPKType.insert(id="b")
    db.CustomPKName.insert(name="a")
    db.CustomPKName.insert(name="b")
    db.CustomPKMulti.insert(first_name="foo", last_name="bar")
    db.CustomPKMulti.insert(first_name="foo", last_name="baz")
    db.CustomPKMulti.insert(first_name="bar", last_name="baz")

    assert Subscription.last().id == Subscription.all().select(
        orderby=~Subscription.id,
        limitby=(0, 1)
    ).first().id
    assert CustomPKType.last().id == CustomPKType.all().select(
        orderby=~CustomPKType.id,
        limitby=(0, 1)
    ).first().id
    assert CustomPKName.last().name == CustomPKName.all().select(
        orderby=~CustomPKName.name,
        limitby=(0, 1)
    ).first().name
    assert CustomPKMulti.last() == CustomPKMulti.all().select(
        orderby=~CustomPKMulti.first_name|~CustomPKMulti.last_name,
        limitby=(0, 1)
    ).first()



# -*- coding: utf-8 -*-
"""
    emmett.orm.engines.sqlite
    -------------------------

    Provides ORM SQLite engine specific features.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from pydal.adapters.sqlite import SQLite as _SQLite

from . import adapters


@adapters.register_for('sqlite', 'sqlite:memory')
class SQLite(_SQLite):
    def _initialize_(self, do_connect):
        super()._initialize_(do_connect)
        self.driver_args['isolation_level'] = None

    def begin(self, lock_type=None):
        statement = 'BEGIN %s;' % lock_type if lock_type else 'BEGIN;'
        self.execute(statement)

    def delete(self, table, query):
        deleted = (
            [x[table._id.name] for x in self.db(query).select(table._id)]
            if table._id else []
        )
        counter = super(_SQLite, self).delete(table, query)
        if table._id and counter:
            for field in table._referenced_by:
                if (
                    field.type == 'reference ' + table._dalname and
                    field.ondelete == 'CASCADE'
                ):
                    self.db(field.belongs(deleted)).delete()
        return counter



# -*- coding: utf-8 -*-
"""
    tests.orm_pks
    -------------

    Test ORM primary keys hendling
"""

import os
import pytest

from uuid import uuid4

from emmett import App, sdict
from emmett.orm import Database, Model, Field, belongs_to, has_many
from emmett.orm.errors import SaveException
from emmett.orm.helpers import RowReferenceMixin
from emmett.orm.migrations.utils import generate_runtime_migration

require_postgres = pytest.mark.skipif(
    not os.environ.get("POSTGRES_URI"), reason="No postgres database"
)


class Standard(Model):
    foo = Field.string()
    bar = Field.string()


class CustomType(Model):
    id = Field.string()
    foo = Field.string()
    bar = Field.string()


class CustomName(Model):
    primary_keys = ["foo"]
    foo = Field.string()
    bar = Field.string()


class CustomMulti(Model):
    primary_keys = ["foo", "bar"]
    foo = Field.string()
    bar = Field.string()
    baz = Field.string()


class SourceCustom(Model):
    has_many("dest_custom_customs", "dest_custom_multis")

    id = Field.string(default=lambda: uuid4().hex)
    foo = Field.string()


class SourceMulti(Model):
    primary_keys = ["foo", "bar"]

    has_many("dest_multi_customs", "dest_multi_multis")

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    baz = Field.string()


class DestCustomCustom(Model):
    belongs_to("source_custom")

    id = Field.string(default=lambda: uuid4().hex)
    foo = Field.string()


class DestCustomMulti(Model):
    primary_keys = ["foo", "bar"]

    belongs_to("source_custom")

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    baz = Field.string()


class DestMultiCustom(Model):
    belongs_to("source_multi")

    id = Field.string(default=lambda: uuid4().hex)
    foo = Field.string()


class DestMultiMulti(Model):
    primary_keys = ["foo", "bar"]

    belongs_to("source_multi")

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    baz = Field.string()


class DoctorCustom(Model):
    has_many(
        {"appointments": "AppointmentCustom"},
        {"patients": {"via": "appointments.patient_custom"}},
        {"symptoms_to_treat": {"via": "patients.symptoms"}}
    )

    id = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class DoctorMulti(Model):
    primary_keys = ["foo", "bar"]

    has_many(
        {"appointments": "AppointmentMulti"},
        {"patients": {"via": "appointments.patient_multi"}},
        {"symptoms_to_treat": {"via": "patients.symptoms"}}
    )

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class PatientCustom(Model):
    primary_keys = ["code"]

    has_many(
        {"appointments": "AppointmentCustom"},
        {"symptoms": "SymptomCustom.patient"},
        {"doctors": {"via": "appointments.doctor_custom"}}
    )

    code = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class PatientMulti(Model):
    primary_keys = ["foo", "bar"]

    has_many(
        {"appointments": "AppointmentMulti"},
        {"symptoms": "SymptomMulti.patient"},
        {"doctors": {"via": "appointments.doctor_multi"}}
    )

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class SymptomCustom(Model):
    belongs_to({"patient": "PatientCustom"})

    id = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class SymptomMulti(Model):
    primary_keys = ["foo", "bar"]

    belongs_to({"patient": "PatientMulti"})

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class AppointmentCustom(Model):
    primary_keys = ["code"]

    belongs_to("patient_custom", "doctor_custom")

    code = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


class AppointmentMulti(Model):
    primary_keys = ["foo", "bar"]

    belongs_to("patient_multi", "doctor_multi")

    foo = Field.string(default=lambda: uuid4().hex)
    bar = Field.string(default=lambda: uuid4().hex)
    name = Field.string()


@pytest.fixture(scope='module')
def _db():
    app = App(__name__)
    db = Database(
        app,
        config=sdict(
            uri=f'sqlite://{uuid4().hex}.db',
            auto_connect=True
        )
    )
    db.define_models(
        Standard,
        CustomType,
        CustomName,
        CustomMulti
    )
    return db


@pytest.fixture(scope='module')
def _pgs():
    app = App(__name__)
    db = Database(
        app,
        config=sdict(
            uri=f"postgres://{os.environ.get('POSTGRES_URI')}",
            auto_connect=True
        )
    )
    db.define_models(
        SourceCustom,
        SourceMulti,
        DestCustomCustom,
        DestCustomMulti,
        DestMultiCustom,
        DestMultiMulti,
        DoctorCustom,
        PatientCustom,
        AppointmentCustom,
        DoctorMulti,
        PatientMulti,
        AppointmentMulti,
        SymptomCustom,
        SymptomMulti
    )
    return db


@pytest.fixture(scope='function')
def db(_db):
    migration = generate_runtime_migration(_db)
    migration.up()
    yield _db
    migration.down()


@pytest.fixture(scope='function')
def pgs(_pgs):
    migration = generate_runtime_migration(_pgs)
    migration.up()
    yield _pgs
    migration.down()


def test_insert(db):
    res = db.Standard.insert(foo="test1", bar="test2")
    assert isinstance(res, int)
    assert res.id
    assert res.foo == "test1"
    assert res.bar == "test2"

    res = db.CustomType.insert(id="test1", foo="test2", bar="test3")
    assert isinstance(res, str)
    assert res.id == "test1"
    assert res.foo == "test2"
    assert res.bar == "test3"

    res = db.CustomName.insert(foo="test1", bar="test2")
    assert isinstance(res, str)
    assert not res.id
    assert res.foo == "test1"
    assert res.bar == "test2"

    res = db.CustomMulti.insert(foo="test1", bar="test2", baz="test3")
    assert isinstance(res, tuple)
    assert not res.id
    assert res.foo == "test1"
    assert res.bar == "test2"
    assert res.baz == "test3"


def test_save_insert(db):
    row = Standard.new(foo="test1", bar="test2")
    done = row.save()
    assert done
    assert row._concrete
    assert row.id
    assert type(row.id) == int

    row = CustomType.new(id="test1", foo="test2", bar="test3")
    done = row.save()
    assert done
    assert row._concrete
    assert row.id == "test1"

    row = CustomName.new(foo="test1", bar="test2")
    done = row.save()
    assert done
    assert row._concrete
    assert "id" not in row
    assert row.foo == "test1"

    row = CustomMulti.new(foo="test1", bar="test2", baz="test3")
    done = row.save()
    assert done
    assert row._concrete
    assert "id" not in row
    assert row.foo == "test1"
    assert row.bar == "test2"
    assert row.baz == "test3"


def test_save_update(db):
    row = Standard.new(foo="test1", bar="test2")
    row.save()
    row.bar = "test2a"
    done = row.save()
    assert done
    assert row._concrete
    assert row.bar == "test2a"
    row.id = 123
    done = row.save()
    assert not done
    with pytest.raises(SaveException):
        row.save(raise_on_error=True)

    row = CustomType.new(id="test1", foo="test2", bar="test3")
    row.save()
    row.bar = "test2a"
    done = row.save()
    assert done
    assert row._concrete
    assert row.bar == "test2a"
    row.id = "test1a"
    done = row.save()
    assert not done
    with pytest.raises(SaveException):
        row.save(raise_on_error=True)

    row = CustomName.new(foo="test1", bar="test2")
    row.save()
    row.bar = "test2a"
    done = row.save()
    assert done
    assert row._concrete
    assert row.bar == "test2a"
    row.foo = "test1a"
    done = row.save()
    assert not done
    with pytest.raises(SaveException):
        row.save(raise_on_error=True)

    row = CustomMulti.new(foo="test1", bar="test2", baz="test3")
    row.save()
    row.baz = "test3a"
    done = row.save()
    assert done
    assert row._concrete
    assert row.baz == "test3a"
    row.foo = "test1a"
    done = row.save()
    assert not done
    with pytest.raises(SaveException):
        row.save(raise_on_error=True)


def test_destroy_delete(db):
    row = Standard.new(foo="test1", bar="test2")
    row.save()
    done = row.destroy()
    assert done
    assert not row._concrete
    assert row.id is None
    assert row.foo == "test1"

    row = CustomType.new(id="test1", foo="test2", bar="test3")
    row.save()
    done = row.destroy()
    assert done
    assert not row._concrete
    assert row.id is None
    assert row.foo == "test2"

    row = CustomName.new(foo="test1", bar="test2")
    row.save()
    done = row.destroy()
    assert done
    assert not row._concrete
    assert row.foo is None
    assert row.bar == "test2"

    row = CustomMulti.new(foo="test1", bar="test2", baz="test3")
    row.save()
    done = row.destroy()
    assert done
    assert not row._concrete
    assert row.foo is None
    assert row.bar is None
    assert row.baz == "test3"


@require_postgres
def test_relations(pgs):
    sc1 = SourceCustom.new(foo="test1")
    sc1.save()
    sc2 = SourceCustom.new(foo="test2")
    sc2.save()
    sm1 = SourceMulti.new(baz="test1")
    sm1.save()
    sm2 = SourceMulti.new(baz="test2")
    sm2.save()

    #: new
    dcc1 = sc1.dest_custom_customs.new(foo="test")
    assert dcc1.source_custom == sc1.id
    assert isinstance(dcc1.source_custom, str)

    dcm1 = sc1.dest_custom_multis.new(baz="test")
    assert dcm1.source_custom == sc1.id
    assert isinstance(dcc1.source_custom, str)

    dmc1 = sm1.dest_multi_customs.new(foo="test")
    assert dmc1.source_multi_foo == sm1.foo
    assert dmc1.source_multi_bar == sm1.bar

    dmm1 = sm1.dest_multi_multis.new(baz="test")
    assert dmm1.source_multi_foo == sm1.foo
    assert dmm1.source_multi_bar == sm1.bar

    #: create
    dcc1 = sc1.dest_custom_customs.create(foo="test")
    assert isinstance(dcc1.id, str)
    row = sc1.dest_custom_customs().first()
    assert row.foo == "test"
    rc = DestCustomCustom.get(row.id)
    assert rc.foo == row.foo
    assert isinstance(rc.source_custom, str)

    dcm1 = sc1.dest_custom_multis.create(baz="test")
    assert isinstance(dcm1.id, tuple)
    row = sc1.dest_custom_multis().first()
    assert row.baz == "test"
    rc = DestCustomMulti.get(row.foo, row.bar)
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_custom, str)
    rc = DestCustomMulti.get(foo=row.foo, bar=row.bar)
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_custom, str)
    rc = DestCustomMulti.get((row.foo, row.bar))
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_custom, str)
    rc = DestCustomMulti.get({"foo": row.foo, "bar": row.bar})
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_custom, str)

    dmc1 = sm1.dest_multi_customs.create(foo="test")
    assert isinstance(dmc1.id, str)
    row = sm1.dest_multi_customs().first()
    assert row.foo == "test"
    rc = DestMultiCustom.get(row.id)
    assert rc.foo == row.foo
    assert isinstance(rc.source_multi, tuple)
    assert rc.source_multi.foo == rc.source_multi_foo
    assert rc.source_multi.bar == rc.source_multi_bar
    assert rc.source_multi.baz == "test1"

    dmm1 = sm1.dest_multi_multis.create(baz="test")
    assert isinstance(dmm1.id, tuple)
    row = sm1.dest_multi_multis().first()
    assert row.baz == "test"
    rc = DestMultiMulti.get(row.foo, row.bar)
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_multi, tuple)
    assert rc.source_multi.foo == rc.source_multi_foo
    assert rc.source_multi.bar == rc.source_multi_bar
    assert rc.source_multi.baz == "test1"
    rc = DestMultiMulti.get(foo=row.foo, bar=row.bar)
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_multi, tuple)
    assert rc.source_multi.foo == rc.source_multi_foo
    assert rc.source_multi.bar == rc.source_multi_bar
    assert rc.source_multi.baz == "test1"
    rc = DestMultiMulti.get((row.foo, row.bar))
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_multi, tuple)
    assert rc.source_multi.foo == rc.source_multi_foo
    assert rc.source_multi.bar == rc.source_multi_bar
    assert rc.source_multi.baz == "test1"
    rc = DestMultiMulti.get({"foo": row.foo, "bar": row.bar})
    assert rc.foo == row.foo
    assert rc.bar == row.bar
    assert isinstance(rc.source_multi, tuple)
    assert rc.source_multi.foo == rc.source_multi_foo
    assert rc.source_multi.bar == rc.source_multi_bar
    assert rc.source_multi.baz == "test1"

    #: add, remove
    dcc1 = DestCustomCustom.first()
    sc2.dest_custom_customs.add(dcc1)
    assert sc1.dest_custom_customs.count() == 0
    assert sc2.dest_custom_customs.count() == 1
    assert dcc1.source_custom.id == sc2.id
    sc2.dest_custom_customs.remove(dcc1)
    assert sc1.dest_custom_customs.count() == 0
    assert sc2.dest_custom_customs.count() == 0
    assert dcc1.source_custom is None
    assert not dcc1.is_valid

    dcm1 = DestCustomMulti.first()
    sc2.dest_custom_multis.add(dcm1)
    assert sc1.dest_custom_multis.count() == 0
    assert sc2.dest_custom_multis.count() == 1
    assert dcm1.source_custom.id == sc2.id
    sc2.dest_custom_multis.remove(dcm1)
    assert sc1.dest_custom_multis.count() == 0
    assert sc2.dest_custom_multis.count() == 0
    assert dcm1.source_custom is None
    assert not dcm1.is_valid

    dmc1 = DestMultiCustom.first()
    sm2.dest_multi_customs.add(dmc1)
    assert sm1.dest_multi_customs.count() == 0
    assert sm2.dest_multi_customs.count() == 1
    assert dmc1.source_multi.foo == sm2.foo
    assert dmc1.source_multi.bar == sm2.bar
    sm2.dest_multi_customs.remove(dmc1)
    assert sm1.dest_multi_customs.count() == 0
    assert sm2.dest_multi_customs.count() == 0
    assert dmc1.source_multi is None
    assert not dmc1.is_valid

    dmm1 = DestMultiMulti.first()
    sm2.dest_multi_multis.add(dmm1)
    assert sm1.dest_multi_multis.count() == 0
    assert sm2.dest_multi_multis.count() == 1
    assert dmm1.source_multi.foo == sm2.foo
    assert dmm1.source_multi.bar == sm2.bar
    sm2.dest_multi_multis.remove(dmm1)
    assert sm1.dest_multi_multis.count() == 0
    assert sm2.dest_multi_multis.count() == 0
    assert dmm1.source_multi is None
    assert not dmm1.is_valid


@require_postgres
def test_via_relations(pgs):
    doc1 = DoctorCustom.new(name="test1")
    doc1.save()
    doc2 = DoctorCustom.new(name="test2")
    doc2.save()
    pat1 = PatientCustom.new(name="test1")
    pat1.save()
    pat1.symptoms.create(name="test1a")
    pat2 = PatientCustom.new(name="test2")
    pat2.save()
    pat2.symptoms.create(name="test2a")
    pat2.symptoms.create(name="test2b")
    doc3 = DoctorMulti.new(name="test1")
    doc3.save()
    doc4 = DoctorMulti.new(name="test2")
    doc4.save()
    pat3 = PatientMulti.new(name="test1")
    pat3.save()
    pat3.symptoms.create(name="test3a")
    pat3.symptoms.create(name="test3b")
    pat4 = PatientMulti.new(name="test2")
    pat4.save()
    pat4.symptoms.create(name="test4a")

    #: add, remove
    doc1.patients.add(pat1, name="test1")
    doc2.patients.add(pat2, name="test2")
    assert doc1.patients.count() == 1
    assert doc1.patients.count() == 1
    assert doc1.symptoms_to_treat.count() == 1
    assert doc2.symptoms_to_treat.count() == 2
    doc1.patients.add(pat2, name="test2")
    assert doc1.patients.count() == 2
    assert doc2.patients.count() == 1
    assert doc1.symptoms_to_treat.count() == 3
    assert doc2.symptoms_to_treat.count() == 2
    doc2.patients.remove(pat2)
    assert doc1.patients.count() == 2
    assert doc2.patients.count() == 0
    assert doc1.symptoms_to_treat.count() == 3
    assert doc2.symptoms_to_treat.count() == 0

    doc3.patients.add(pat3, name="test1")
    doc4.patients.add(pat4, name="test2")
    assert doc3.patients.count() == 1
    assert doc3.patients.count() == 1
    assert doc3.symptoms_to_treat.count() == 2
    assert doc4.symptoms_to_treat.count() == 1
    doc3.patients.add(pat4, name="test2")
    assert doc3.patients.count() == 2
    assert doc4.patients.count() == 1
    assert doc3.symptoms_to_treat.count() == 3
    assert doc4.symptoms_to_treat.count() == 1
    doc4.patients.remove(pat4)
    assert doc3.patients.count() == 2
    assert doc4.patients.count() == 0
    assert doc3.symptoms_to_treat.count() == 3
    assert doc4.symptoms_to_treat.count() == 0


@require_postgres
def test_relations_set(pgs):
    doc1 = DoctorCustom.new(name="test1")
    doc1.save()
    doc2 = DoctorCustom.new(name="test2")
    doc2.save()
    pat1 = PatientCustom.new(name="test1")
    pat1.save()
    pat1.symptoms.create(name="test1a")
    pat2 = PatientCustom.new(name="test2")
    pat2.save()
    pat2.symptoms.create(name="test2a")
    pat2.symptoms.create(name="test2b")
    doc3 = DoctorMulti.new(name="test1")
    doc3.save()
    doc4 = DoctorMulti.new(name="test2")
    doc4.save()
    pat3 = PatientMulti.new(name="test1")
    pat3.save()
    pat3.symptoms.create(name="test3a")
    pat3.symptoms.create(name="test3b")
    pat4 = PatientMulti.new(name="test2")
    pat4.save()
    pat4.symptoms.create(name="test4a")

    doc1.patients.add(pat1, name="test1")

    djoin = DoctorCustom.all().join("appointments").select()
    assert len(djoin) == 1
    assert djoin[0].id == doc1.id
    assert len(djoin[0].appointments()) == 1

    djoin = DoctorCustom.all().join("patients").select()
    assert len(djoin) == 1
    assert djoin[0].id == doc1.id
    assert len(djoin[0].patients()) == 1
    assert djoin[0].patients()[0].code == pat1.code

    pjoin = PatientCustom.all().join("appointments").select()
    assert len(pjoin) == 1
    assert pjoin[0].code == pat1.code
    assert len(pjoin[0].appointments()) == 1

    pjoin = PatientCustom.all().join("doctors").select()
    assert len(pjoin) == 1
    assert pjoin[0].code == pat1.code
    assert len(pjoin[0].doctors()) == 1
    assert pjoin[0].doctors()[0].id == doc1.id

    ajoin = AppointmentCustom.all().join("doctor_custom", "patient_custom").select()
    assert len(ajoin) == 1
    assert ajoin[0].doctor_custom.id == doc1.id
    assert ajoin[0].patient_custom.code == pat1.code

    djoin = DoctorCustom.all().select(including=["appointments"])
    assert len(djoin) == 2
    assert len(djoin[0].appointments()) == 1
    assert len(djoin[1].appointments()) == 0

    djoin = DoctorCustom.all().join("appointments").select(including=["patients"])
    assert len(djoin) == 1
    assert djoin[0].id == doc1.id
    assert len(djoin[0].patients()) == 1

    pjoin = PatientCustom.all().select(including=["appointments"])
    assert len(pjoin) == 2
    assert len(pjoin[0].appointments()) == 1
    assert len(pjoin[1].appointments()) == 0

    pjoin = PatientCustom.all().join("appointments").select(including=["doctors"])
    assert len(pjoin) == 1
    assert pjoin[0].code == pat1.code
    assert len(pjoin[0].doctors()) == 1

    doc3.patients.add(pat3, name="test1")

    djoin = DoctorMulti.all().join("appointments").select()
    assert len(djoin) == 1
    assert djoin[0].foo == doc3.foo
    assert djoin[0].bar == doc3.bar
    assert len(djoin[0].appointments()) == 1

    djoin = DoctorMulti.all().join("patients").select()
    assert len(djoin) == 1
    assert djoin[0].foo == doc3.foo
    assert djoin[0].bar == doc3.bar
    assert len(djoin[0].patients()) == 1
    assert djoin[0].patients()[0].foo == pat3.foo
    assert djoin[0].patients()[0].bar == pat3.bar

    pjoin = PatientMulti.all().join("appointments").select()
    assert len(pjoin) == 1
    assert pjoin[0].foo == pat3.foo
    assert pjoin[0].bar == pat3.bar
    assert len(pjoin[0].appointments()) == 1

    pjoin = PatientMulti.all().join("doctors").select()
    assert len(pjoin) == 1
    assert pjoin[0].foo == pat3.foo
    assert pjoin[0].bar == pat3.bar
    assert len(pjoin[0].doctors()) == 1
    assert pjoin[0].doctors()[0].foo == doc3.foo
    assert pjoin[0].doctors()[0].bar == doc3.bar

    ajoin = AppointmentMulti.all().join("doctor_multi", "patient_multi").select()
    assert len(ajoin) == 1
    assert ajoin[0].doctor_multi.foo == doc3.foo
    assert ajoin[0].doctor_multi.bar == doc3.bar
    assert ajoin[0].patient_multi.foo == pat3.foo
    assert ajoin[0].patient_multi.bar == pat3.bar

    djoin = DoctorMulti.all().select(including=["appointments"])
    assert len(djoin) == 2
    assert len(djoin[0].appointments()) == 1
    assert len(djoin[1].appointments()) == 0

    djoin = DoctorMulti.all().join("appointments").select(including=["patients"])
    assert len(djoin) == 1
    assert djoin[0].foo == doc3.foo
    assert djoin[0].bar == doc3.bar
    assert len(djoin[0].patients()) == 1

    pjoin = PatientMulti.all().select(including=["appointments"])
    assert len(pjoin) == 2
    assert len(pjoin[0].appointments()) == 1
    assert len(pjoin[1].appointments()) == 0

    pjoin = PatientMulti.all().join("appointments").select(including=["doctors"])
    assert len(pjoin) == 1
    assert pjoin[0].foo == pat3.foo
    assert pjoin[0].bar == pat3.bar
    assert len(pjoin[0].doctors()) == 1


@require_postgres
def test_row(pgs):
    sc1 = SourceCustom.new(foo="test1")
    sc1.save()
    sc2 = SourceCustom.new(foo="test2")
    sc2.save()

    dcc1 = DestCustomCustom.new(foo="test1", source_custom=sc1.id)
    assert isinstance(dcc1.source_custom, RowReferenceMixin)
    dcc1.save()
    assert isinstance(dcc1.source_custom, RowReferenceMixin)

    dcc1 = DestCustomCustom.get(dcc1.id)
    dcc1.source_custom = sc2.id
    assert isinstance(dcc1.source_custom, RowReferenceMixin)
    dcc1.save()
    assert isinstance(dcc1.source_custom, RowReferenceMixin)

    sm1 = SourceMulti.new(baz="test1")
    sm1.save()
    sm2 = SourceMulti.new(baz="test2")
    sm2.save()

    dmm1 = DestMultiMulti.new(source_multi=sm1, baz="test")
    dmm1.save()
    assert sm1.dest_multi_multis.count() == 1
    assert sm2.dest_multi_multis.count() == 0

    dmm1.source_multi = sm2
    assert set(dmm1._changes.keys()).issubset(
        {"source_multi", "source_multi_foo", "source_multi_bar"}
    )
    dmm1.save()
    assert sm1.dest_multi_multis.count() == 0
    assert sm2.dest_multi_multis.count() == 1

    DestMultiMulti.create(source_multi=sm1, baz="test")
    assert sm1.dest_multi_multis.count() == 1



# -*- coding: utf-8 -*-
"""
    emmett.rsgi.wrappers
    --------------------

    Provides RSGI request and websocket wrappers

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from datetime import datetime
from typing import Any, Dict, List, Union, Optional
from urllib.parse import parse_qs

from granian.rsgi import Scope, HTTPProtocol, ProtocolClosed

from .helpers import WSTransport
from ..datastructures import sdict
from ..http import HTTP
from ..utils import cachedprop
from ..wrappers.helpers import regex_client
from ..wrappers.request import Request as _Request
from ..wrappers.websocket import Websocket as _Websocket


class RSGIIngressMixin:
    def __init__(
        self,
        scope: Scope,
        path: str,
        protocol: Union[HTTPProtocol, WSTransport]
    ):
        self._scope = scope
        self._proto = protocol
        self.scheme = scope.scheme
        self.path = path

    @property
    def headers(self):
        return self._scope.headers

    @cachedprop
    def query_params(self) -> sdict[str, Union[str, List[str]]]:
        rv: sdict[str, Any] = sdict()
        for key, values in parse_qs(
            self._scope.query_string, keep_blank_values=True
        ).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv


class Request(RSGIIngressMixin, _Request):
    __slots__ = ['_scope', '_proto']

    def __init__(
        self,
        scope: Scope,
        path: str,
        protocol: HTTPProtocol,
        max_content_length: Optional[int] = None,
        body_timeout: Optional[int] = None
    ):
        super().__init__(scope, path, protocol)
        self.max_content_length = max_content_length
        self.body_timeout = body_timeout
        self._now = datetime.utcnow()
        self.method = scope.method

    @property
    def _multipart_headers(self):
        return dict(self.headers.items())

    @cachedprop
    async def body(self) -> bytes:
        if (
            self.max_content_length and
            self.content_length > self.max_content_length
        ):
            raise HTTP(413, 'Request entity too large')
        try:
            rv = await asyncio.wait_for(self._proto(), timeout=self.body_timeout)
        except asyncio.TimeoutError:
            raise HTTP(408, 'Request timeout')
        return rv

    @cachedprop
    def client(self) -> str:
        g = regex_client.search(self.headers.get('x-forwarded-for', ''))
        client = (
            (g.group() or '').split(',')[0] if g else (
                self._scope.client[0] if self._scope.client else None
            )
        )
        if client in (None, '', 'unknown', 'localhost'):
            client = '::1' if self.host.startswith('[') else '127.0.0.1'
        return client  # type: ignore

    async def push_promise(self, path: str):
        raise NotImplementedError("RSGI protocol doesn't support HTTP2 push.")


class Websocket(RSGIIngressMixin, _Websocket):
    __slots__ = ['_scope', '_proto']

    def __init__(
        self,
        scope: Scope,
        path: str,
        protocol: WSTransport
    ):
        super().__init__(scope, path, protocol)
        self._flow_receive = None
        self._flow_send = None
        self.receive = self._accept_and_receive
        self.send = self._accept_and_send

    async def accept(
        self,
        headers: Optional[Dict[str, str]] = None,
        subprotocol: Optional[str] = None
    ):
        if self._proto.transport:
            return
        await self._proto.init()
        self.receive = self._wrapped_receive
        self.send = self._wrapped_send

    async def _wrapped_receive(self) -> Any:
        data = (await self._proto.receive()).data
        for method in self._flow_receive:
            data = method(data)
        return data

    async def _wrapped_send(self, data: Any):
        for method in self._flow_send:
            data = method(data)
        trx = (
            self._proto.transport.send_str if isinstance(data, str) else
            self._proto.transport.send_bytes
        )
        try:
            await trx(data)
        except ProtocolClosed:
            if not self._proto.interrupted:
                raise
            await self._proto.noop.wait()



# -*- coding: utf-8 -*-
"""
    emmett.orm.geo
    --------------

    Provides geographic facilities.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from .helpers import GeoFieldWrapper


def Point(x, y):
    return GeoFieldWrapper("POINT(%f %f)" % (x, y))


def Line(*coordinates):
    return GeoFieldWrapper(
        "LINESTRING(%s)" % ','.join("%f %f" % point for point in coordinates)
    )


def Polygon(*coordinates_groups):
    try:
        if not isinstance(coordinates_groups[0][0], (tuple, list)):
            coordinates_groups = (coordinates_groups,)
    except Exception:
        pass
    return GeoFieldWrapper(
        "POLYGON(%s)" % (
            ",".join([
                "(%s)" % ",".join("%f %f" % point for point in group)
                for group in coordinates_groups
            ])
        )
    )


def MultiPoint(*points):
    return GeoFieldWrapper(
        "MULTIPOINT(%s)" % (
            ",".join([
                "(%f %f)" % point for point in points
            ])
        )
    )


def MultiLine(*lines):
    return GeoFieldWrapper(
        "MULTILINESTRING(%s)" % (
            ",".join([
                "(%s)" % ",".join("%f %f" % point for point in line)
                for line in lines
            ])
        )
    )


def MultiPolygon(*polygons):
    return GeoFieldWrapper(
        "MULTIPOLYGON(%s)" % (
            ",".join([
                "(%s)" % (
                    ",".join([
                        "(%s)" % ",".join("%f %f" % point for point in group)
                        for group in polygon
                    ])
                ) for polygon in polygons
            ])
        )
    )



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.scripts
    -----------------------------

    Provides scripts interface for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

import os
import re
import sys

from contextlib import contextmanager
from datetime import datetime
from importlib import resources

from renoir import Renoir

from ...html import asis
from . import __name__ as __pkg__
from .base import Migration
from .exceptions import (
    RangeNotAncestorError, MultipleHeads, ResolutionError, RevisionError
)
from .helpers import tuple_rev_as_scalar, format_with_comma
from .revisions import Revision, RevisionsMap


class ScriptDir(object):
    _slug_re = re.compile(r'\w+')
    _default_file_template = "%(rev)s_%(slug)s"

    def __init__(self, app, migrations_folder=None):
        self.app = app
        self.path = os.path.join(
            app.root_path, migrations_folder or 'migrations')
        if not os.path.exists(self.path):
            os.mkdir(self.path)
        self.cwd = os.path.dirname(__file__)
        self.file_template = self.app.config.migrations.file_template or \
            self._default_file_template
        self.truncate_slug_length = \
            self.app.config.migrations.filename_len or 40
        self.revision_map = RevisionsMap(self.app, self._load_revisions)
        self.templater = Renoir(path=self.cwd, mode='plain')

    def _load_revisions(self):
        sys.path.insert(0, self.path)
        for rev_file in os.listdir(os.path.abspath(self.path)):
            script = Script._from_filename(self, rev_file)
            if script is None:
                continue
            yield script

    @contextmanager
    def _catch_revision_errors(
            self,
            ancestor=None, multiple_heads=None, start=None, end=None,
            resolution=None):
        try:
            yield
        except RangeNotAncestorError as rna:
            if start is None:
                start = rna.lower
            if end is None:
                end = rna.upper
            if not ancestor:
                ancestor = (
                    "Requested range %(start)s:%(end)s does not refer to "
                    "ancestor/descendant revisions along the same branch"
                )
            ancestor = ancestor % {"start": start, "end": end}
            raise Exception(ancestor)
        except MultipleHeads as mh:
            if not multiple_heads:
                multiple_heads = (
                    "Multiple head revisions are present for given "
                    "argument '%(head_arg)s'; please "
                    "specify a specific target revision, "
                    "'<branchname>@%(head_arg)s' to "
                    "narrow to a specific head, or 'heads' for all heads")
            multiple_heads = multiple_heads % {
                "head_arg": end or mh.argument,
                "heads": str(mh.heads)
            }
            raise Exception(multiple_heads)
        except ResolutionError as re:
            if resolution is None:
                resolution = "Can't locate revision identified by '%s'" % (
                    re.argument
                )
            raise Exception(resolution)
        except RevisionError as err:
            raise Exception(err.args[0])

    def walk_revisions(self, base="base", head="heads"):
        with self._catch_revision_errors(start=base, end=head):
            for rev in self.revision_map.iterate_revisions(
                    head, base, inclusive=True):
                yield rev

    def get_revision(self, revid):
        with self._catch_revision_errors():
            return self.revision_map.get_revision(revid)

    def get_revisions(self, revid):
        with self._catch_revision_errors():
            return self.revision_map.get_revisions(revid)

    def get_upgrade_revs(self, destination, current_rev):
        with self._catch_revision_errors(
                ancestor="Destination %(end)s is not a valid upgrade "
                "target from current head(s)", end=destination):
            revs = self.revision_map.iterate_revisions(
                destination, current_rev, implicit_base=True)
            return reversed(list(revs))

    def get_downgrade_revs(self, destination, current_rev):
        with self._catch_revision_errors(
                ancestor="Destination %(end)s is not a valid downgrade "
                "target from current head(s)", end=destination):
            revs = self.revision_map.iterate_revisions(
                current_rev, destination)
            return list(revs)

    def _rev_filename(self, revid, message, creation_date):
        slug = "_".join(self._slug_re.findall(message or "")).lower()
        if len(slug) > self.truncate_slug_length:
            slug = slug[:self.truncate_slug_length].rsplit('_', 1)[0] + '_'
        filename = "%s.py" % (
            self.file_template % {
                'rev': revid,
                'slug': slug,
                'year': creation_date.year,
                'month': creation_date.month,
                'day': creation_date.day,
                'hour': creation_date.hour,
                'minute': creation_date.minute,
                'second': creation_date.second
            }
        )
        return filename

    def _generate_template(self, filename, ctx):
        tmpl_source = resources.read_text(__pkg__, 'migration.tmpl')
        rendered = self.templater._render(source=tmpl_source, context=ctx)
        with open(os.path.join(self.path, filename), 'w') as f:
            f.write(rendered)

    def generate_revision(self, revid, message, head=None, splice=False, **kw):
        """Generate a new revision file.

        This runs the templater, and creates a new file.

        :param revid: String revision id.
        :param message: the revision message.
        :param head: the head revision to generate against. Defaults
         to the current "head" if no branches are present, else raises
         an exception.
        :param splice: if True, allow the "head" version to not be an
         actual head; otherwise, the selected head must be a head
         (e.g. endpoint) revision.
        """
        if head is None:
            head = "head"

        with self._catch_revision_errors(multiple_heads=(
            "Multiple heads are present; please specify the head "
            "revision on which the new revision should be based, "
            "or perform a merge."
        )):
            heads = self.revision_map.get_revisions(head)

        if len(set(heads)) != len(heads):
            raise Exception("Duplicate head revisions specified")

        creation_date = datetime.now()

        rev_filename = self._rev_filename(revid, message, creation_date)

        if not splice:
            for head in heads:
                if head is not None and not head.is_head:
                    raise Exception(
                        "Revision %s is not a head revision; please specify "
                        "--splice to create a new branch from this revision"
                        % head.revision)

        down_migration = tuple(
            h.revision if h is not None else None for h in heads)

        down_migration_var = tuple_rev_as_scalar(down_migration)
        if isinstance(down_migration_var, str):
            down_migration_var = "%r" % down_migration_var
        else:
            down_migration_var = str(down_migration_var)

        template_ctx = dict(
            asis=asis,
            up_migration=revid,
            down_migration=down_migration_var,
            creation_date=creation_date,
            down_migration_str=", ".join(r for r in down_migration),
            message=message if message is not None else ("empty message"),
            upgrades=kw.get('upgrades', ['pass']),
            downgrades=kw.get('downgrades', ['pass'])
        )
        self._generate_template(rev_filename, template_ctx)

        script = Script._from_filename(self, rev_filename)
        self.revision_map.add_revision(script)
        return script


class Script(Revision):
    _only_source_rev_file = re.compile(r'(?!__init__)(.*\.py)$')
    migration_class = None
    path = None

    def __init__(self, module, migration_class, path):
        self.module = module
        self.migration_class = migration_class
        self.path = path
        super(Script, self).__init__(
            self.migration_class.revision,
            self.migration_class.revises
        )

    @property
    def doc(self):
        return re.split("\n\n", self.longdoc)[0]

    @property
    def longdoc(self):
        doc = self.module.__doc__
        return doc.strip() if doc else ""

    @property
    def log_entry(self):
        entry = "Rev: %s%s%s%s\n" % (
            self.revision,
            " (head)" if self.is_head else "",
            " (branchpoint)" if self.is_branch_point else "",
            " (mergepoint)" if self.is_merge_point else "",
        )
        if self.is_merge_point:
            entry += "Merges: %s\n" % (self._format_down_revision(), )
        else:
            entry += "Parent: %s\n" % (self._format_down_revision(), )

        if self.is_branch_point:
            entry += "Branches into: %s\n" % (
                format_with_comma(self.nextrev))

        entry += "Path: %s\n" % (self.path,)

        entry += "\n%s\n" % (
            "\n".join(
                "    %s" % para
                for para in self.longdoc.splitlines()
            )
        )
        return entry

    def __str__(self):
        return "%s -> %s%s%s%s, %s" % (
            self._format_down_revision(),
            self.revision,
            " (head)" if self.is_head else "",
            " (branchpoint)" if self.is_branch_point else "",
            " (mergepoint)" if self.is_merge_point else "",
            self.doc)

    def _head_only(
            self, include_doc=False,
            include_parents=False, tree_indicators=True,
            head_indicators=True):
        text = self.revision
        if include_parents:
            text = "%s -> %s" % (
                self._format_down_revision(), text)
        if head_indicators or tree_indicators:
            text += "%s%s" % (
                " (head)" if self._is_real_head else "",
                " (effective head)" if self.is_head and
                    not self._is_real_head else ""
            )
        if tree_indicators:
            text += "%s%s" % (
                " (branchpoint)" if self.is_branch_point else "",
                " (mergepoint)" if self.is_merge_point else "",
            )
        if include_doc:
            text += ", %s" % self.doc
        return text

    def cmd_format(
        self,
            verbose,
            include_doc=False,
            include_parents=False, tree_indicators=True):
        if verbose:
            return self.log_entry
        else:
            return self._head_only(
                include_doc,
                include_parents, tree_indicators)

    def _format_down_revision(self):
        if not self.down_revision:
            return "<base>"
        else:
            return format_with_comma(self._versioned_down_revisions)

    @classmethod
    def _from_filename(cls, scriptdir, filename):
        py_match = cls._only_source_rev_file.match(filename)
        if not py_match:
            return None
        py_filename = py_match.group(1)
        py_module = py_filename.split('.py')[0]
        __import__(py_module)
        module = sys.modules[py_module]
        migration_class = getattr(module, 'Migration', None)
        if migration_class is None:
            for v in module.__dict__.values():
                if isinstance(v, Migration):
                    migration_class = v
                    break
        return Script(
            module, migration_class, os.path.join(scriptdir.path, filename))



# -*- coding: utf-8 -*-
"""
    emmett.http
    -----------

    Provides the HTTP interfaces.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import errno
import os
import stat

from email.utils import formatdate
from hashlib import md5
from typing import Any, AsyncIterable, BinaryIO, Dict, Generator, Iterable, Tuple

from granian.rsgi import HTTPProtocol

from ._internal import loop_open_file
from .ctx import current
from .libs.contenttype import contenttype


status_codes = {
    100: '100 CONTINUE',
    101: '101 SWITCHING PROTOCOLS',
    200: '200 OK',
    201: '201 CREATED',
    202: '202 ACCEPTED',
    203: '203 NON-AUTHORITATIVE INFORMATION',
    204: '204 NO CONTENT',
    205: '205 RESET CONTENT',
    206: '206 PARTIAL CONTENT',
    207: '207 MULTI-STATUS',
    300: '300 MULTIPLE CHOICES',
    301: '301 MOVED PERMANENTLY',
    302: '302 FOUND',
    303: '303 SEE OTHER',
    304: '304 NOT MODIFIED',
    305: '305 USE PROXY',
    307: '307 TEMPORARY REDIRECT',
    400: '400 BAD REQUEST',
    401: '401 UNAUTHORIZED',
    403: '403 FORBIDDEN',
    404: '404 NOT FOUND',
    405: '405 METHOD NOT ALLOWED',
    406: '406 NOT ACCEPTABLE',
    407: '407 PROXY AUTHENTICATION REQUIRED',
    408: '408 REQUEST TIMEOUT',
    409: '409 CONFLICT',
    410: '410 GONE',
    411: '411 LENGTH REQUIRED',
    412: '412 PRECONDITION FAILED',
    413: '413 REQUEST ENTITY TOO LARGE',
    414: '414 REQUEST-URI TOO LONG',
    415: '415 UNSUPPORTED MEDIA TYPE',
    416: '416 REQUESTED RANGE NOT SATISFIABLE',
    417: '417 EXPECTATION FAILED',
    422: '422 UNPROCESSABLE ENTITY',
    500: '500 INTERNAL SERVER ERROR',
    501: '501 NOT IMPLEMENTED',
    502: '502 BAD GATEWAY',
    503: '503 SERVICE UNAVAILABLE',
    504: '504 GATEWAY TIMEOUT',
    505: '505 HTTP VERSION NOT SUPPORTED',
}


class HTTPResponse(Exception):
    def __init__(
        self,
        status_code: int,
        *,
        headers: Dict[str, str] = {'content-type': 'text/plain'},
        cookies: Dict[str, Any] = {}
    ):
        self.status_code: int = status_code
        self._headers: Dict[str, str] = headers
        self._cookies: Dict[str, Any] = cookies

    @property
    def headers(self) -> Generator[Tuple[bytes, bytes], None, None]:
        for key, val in self._headers.items():
            yield key.encode('latin-1'), val.encode('latin-1')
        for cookie in self._cookies.values():
            yield b'set-cookie', str(cookie)[12:].encode('latin-1')

    @property
    def rsgi_headers(self) -> Generator[Tuple[str, str], None, None]:
        for key, val in self._headers.items():
            yield key, val
        for cookie in self._cookies.values():
            yield 'set-cookie', str(cookie)[12:]

    async def _send_headers(self, send):
        await send({
            'type': 'http.response.start',
            'status': self.status_code,
            'headers': list(self.headers)
        })

    async def _send_body(self, send):
        await send({'type': 'http.response.body'})

    async def asgi(self, scope, send):
        await self._send_headers(send)
        await self._send_body(send)

    def rsgi(self, protocol: HTTPProtocol):
        protocol.response_empty(
            self.status_code,
            list(self.rsgi_headers)
        )


class HTTPBytes(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        body: bytes = b'',
        headers: Dict[str, str] = {'content-type': 'text/plain'},
        cookies: Dict[str, Any] = {}
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.body = body

    async def _send_body(self, send):
        await send({
            'type': 'http.response.body',
            'body': self.body,
            'more_body': False
        })

    def rsgi(self, protocol: HTTPProtocol):
        protocol.response_bytes(
            self.status_code,
            list(self.rsgi_headers),
            self.body
        )


class HTTP(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        body: str = '',
        headers: Dict[str, str] = {'content-type': 'text/plain'},
        cookies: Dict[str, Any] = {}
    ):
        super().__init__(status_code, headers=headers, cookies=cookies)
        self.body = body

    @property
    def encoded_body(self):
        return self.body.encode('utf-8')

    async def _send_body(self, send):
        await send({
            'type': 'http.response.body',
            'body': self.encoded_body,
            'more_body': False
        })

    def rsgi(self, protocol: HTTPProtocol):
        protocol.response_str(
            self.status_code,
            list(self.rsgi_headers),
            self.body
        )


class HTTPRedirect(HTTPResponse):
    def __init__(
        self,
        status_code: int,
        location: str,
        cookies: Dict[str, Any] = {}
    ):
        location = location.replace('\r', '%0D').replace('\n', '%0A')
        super().__init__(
            status_code,
            headers={'location': location},
            cookies=cookies
        )


class HTTPFile(HTTPResponse):
    def __init__(
        self,
        file_path: str,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
        chunk_size: int = 4096
    ):
        super().__init__(200, headers=headers, cookies=cookies)
        self.file_path = file_path
        self.chunk_size = chunk_size

    def _get_stat_headers(self, stat_data):
        content_length = str(stat_data.st_size)
        last_modified = formatdate(stat_data.st_mtime, usegmt=True)
        etag_base = str(stat_data.st_mtime) + '_' + str(stat_data.st_size)
        etag = md5(etag_base.encode('utf-8')).hexdigest()
        return {
            'content-type': contenttype(self.file_path),
            'content-length': content_length,
            'last-modified': last_modified,
            'etag': etag
        }

    async def asgi(self, scope, send):
        try:
            stat_data = os.stat(self.file_path)
            if not stat.S_ISREG(stat_data.st_mode):
                await HTTP(403).send(scope, send)
                return
            self._headers.update(self._get_stat_headers(stat_data))
            await self._send_headers(send)
            await self._send_body(send)
        except IOError as e:
            if e.errno == errno.EACCES:
                await HTTP(403).send(scope, send)
            else:
                await HTTP(404).send(scope, send)

    async def _send_body(self, send):
        async with loop_open_file(self.file_path, mode='rb') as f:
            more_body = True
            while more_body:
                chunk = await f.read(self.chunk_size)
                more_body = len(chunk) == self.chunk_size
                await send({
                    'type': 'http.response.body',
                    'body': chunk,
                    'more_body': more_body,
                })

    def rsgi(self, protocol: HTTPProtocol):
        try:
            stat_data = os.stat(self.file_path)
            if not stat.S_ISREG(stat_data.st_mode):
                return HTTP(403).rsgi(protocol)
            self._headers.update(self._get_stat_headers(stat_data))
        except IOError as e:
            if e.errno == errno.EACCES:
                return HTTP(403).rsgi(protocol)
            return HTTP(404).rsgi(protocol)

        protocol.response_file(
            self.status_code,
            list(self.rsgi_headers),
            self.file_path
        )


class HTTPIO(HTTPResponse):
    def __init__(
        self,
        io_stream: BinaryIO,
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {},
        chunk_size: int = 4096
    ):
        super().__init__(200, headers=headers, cookies=cookies)
        self.io_stream = io_stream
        self.chunk_size = chunk_size

    def _get_io_headers(self):
        content_length = str(self.io_stream.getbuffer().nbytes)
        return {
            'content-length': content_length
        }

    async def asgi(self, scope, send):
        self._headers.update(self._get_io_headers())
        await self._send_headers(send)
        await self._send_body(send)

    async def _send_body(self, send):
        more_body = True
        while more_body:
            chunk = self.io_stream.read(self.chunk_size)
            more_body = len(chunk) == self.chunk_size
            await send({
                'type': 'http.response.body',
                'body': chunk,
                'more_body': more_body,
            })

    def rsgi(self, protocol: HTTPProtocol):
        protocol.response_bytes(
            self.status_code,
            list(self.rsgi_headers),
            self.io_stream.read()
        )


class HTTPIter(HTTPResponse):
    def __init__(
        self,
        iter: Iterable[bytes],
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {}
    ):
        super().__init__(200, headers=headers, cookies=cookies)
        self.iter = iter

    async def _send_body(self, send):
        for chunk in self.iter:
            await send({
                'type': 'http.response.body',
                'body': chunk,
                'more_body': True
            })
        await send({'type': 'http.response.body', 'body': b'', 'more_body': False})

    async def rsgi(self, protocol: HTTPProtocol):
        trx = protocol.response_stream(
            self.status_code,
            list(self.rsgi_headers)
        )
        for chunk in self.iter:
            await trx.send_bytes(chunk)


class HTTPAiter(HTTPResponse):
    def __init__(
        self,
        iter: AsyncIterable[bytes],
        headers: Dict[str, str] = {},
        cookies: Dict[str, Any] = {}
    ):
        super().__init__(200, headers=headers, cookies=cookies)
        self.iter = iter

    async def _send_body(self, send):
        async for chunk in self.iter:
            await send({
                'type': 'http.response.body',
                'body': chunk,
                'more_body': True
            })
        await send({'type': 'http.response.body', 'body': b'', 'more_body': False})

    async def rsgi(self, protocol: HTTPProtocol):
        trx = protocol.response_stream(
            self.status_code,
            list(self.rsgi_headers)
        )
        async for chunk in self.iter:
            await trx.send_bytes(chunk)


def redirect(location: str, status_code: int = 303):
    response = current.response
    response.status = status_code
    raise HTTPRedirect(status_code, location, response.cookies)



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.generation
    --------------------------------

    Provides generation utils for migrations.

    :copyright: 2014 Giovanni Barillari

    Based on the code of Alembic (https://bitbucket.org/zzzeek/alembic)
    :copyright: (c) 2009-2015 by Michael Bayer

    :license: BSD-3-Clause
"""

from __future__ import annotations

from collections import OrderedDict
from typing import Any, Dict, List, Optional

from ..._shortcuts import hashlib_sha1
from ...datastructures import OrderedSet
from ..objects import Rows, Table
from .base import Column, Database
from .helpers import Dispatcher, DEFAULT_VALUE
from .operations import (
    AddColumnOp,
    AlterColumnOp,
    CreateForeignKeyConstraintOp,
    CreateIndexOp,
    CreateTableOp,
    DropColumnOp,
    DropForeignKeyConstraintOp,
    DropIndexOp,
    DropTableOp,
    MigrationOp,
    OpContainer,
    Operation,
    UpgradeOps
)
from .scripts import ScriptDir


class MetaTable:
    def __init__(
        self,
        name: str,
        columns: List[Column] = [],
        primary_keys: List[str] = [],
        **kw: Any
    ):
        self.name = name
        self.columns = OrderedDict()
        for column in columns:
            self.columns[column.name] = column
        self.primary_keys = primary_keys
        self.indexes: Dict[str, MetaIndex] = {}
        self.foreign_keys: Dict[str, MetaForeignKey] = {}
        self.kw = kw

    @property
    def fields(self) -> List[str]:
        return list(self.columns)

    def __getitem__(self, name: str) -> Column:
        return self.columns[name]

    def __setitem__(self, name: str, value: Column):
        self.columns[name] = value

    def __delitem__(self, name: str):
        del self.columns[name]

    def __repr__(self) -> str:
        return "Table(%r, %s)" % (
            self.name,
            ", ".join(["%s" % column for column in self.columns.values()])
        )

    def insert(self, *args, **kwargs) -> Any:
        return None


class MetaIndex:
    def __init__(
        self,
        table_name: str,
        name: str,
        fields: List[str],
        expressions: List[str],
        unique: bool,
        **kw: Any
    ):
        self.table_name = table_name
        self.name = name
        self.fields = fields
        self.expressions = expressions
        self.unique = unique
        self.kw = kw

    @property
    def where(self) -> Optional[str]:
        return self.kw.get('where')

    def __repr__(self) -> str:
        opts = [('expressions', self.expressions), ('unique', self.unique)]
        for key, val in self.kw.items():
            opts.append((key, val))
        return "Index(%r, %r, %s)" % (
            self.name, self.fields,
            ", ".join(["%s=%r" % (opt[0], opt[1]) for opt in opts])
        )


class MetaForeignKey:
    def __init__(
        self,
        table_name: str,
        name: str,
        column_names: List[str],
        foreign_table_name: str,
        foreign_keys: List[str],
        on_delete: str,
        **kw
    ):
        self.table_name = table_name
        self.name = name
        self.column_names = column_names
        self.foreign_table_name = foreign_table_name
        self.foreign_keys = foreign_keys
        self.on_delete = on_delete
        self.kw = kw

    @property
    def _hash(self) -> str:
        return hashlib_sha1(
            f"{self.table_name}:{self.name}:{self.on_delete}:"
            f"{repr(sorted(self.column_names))}:{repr(sorted(self.foreign_keys))}"
        ).hexdigest()

    def __eq__(self, obj: Any) -> bool:
        if isinstance(obj, MetaForeignKey):
            return self._hash == obj._hash
        return False

    def __repr__(self) -> str:
        return "ForeignKey(%r, %r, %r, %r, %r, on_delete=%r)" % (
            self.name,
            self.table_name,
            self.foreign_table_name,
            self.column_names,
            self.foreign_keys,
            self.on_delete
        )


class MetaDataSet:
    def __init__(self, db: MetaData, *args, **kwargs):
        self.db = db
        self.args = args
        self.kwargs = kwargs

    def where(self, *args, **kwargs) -> MetaDataSet:
        return MetaDataSet(self.db, *args, **kwargs)

    def select(self, *args, **kwargs) -> Rows:
        return Rows(self.db)

    def update(self, *args, **kwargs) -> int:
        return 0

    def delete(self, *args, **kwargs) -> int:
        return 0


class MetaData:
    def __init__(self):
        self.tables: Dict[str, MetaTable] = {}

    def __getitem__(self, key):
        return self.tables[key]

    def __getattr__(self, name):
        return self.tables[name]

    def where(self, *args, **kwargs) -> MetaDataSet:
        return MetaDataSet(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.where(*args, **kwargs)

    def create_table(
        self,
        name: str,
        columns: List[Column],
        primary_keys: List[str],
        **kw: Any
    ):
        self.tables[name] = MetaTable(name, columns, primary_keys, **kw)

    def drop_table(self, name: str):
        del self.tables[name]

    def add_column(self, table: str, column: Column):
        self.tables[table][column.name] = column

    def drop_column(self, table: str, column: str):
        del self.tables[table][column]

    def change_column(self, table_name: str, column_name: str, changes: Dict[str, Any]):
        self.tables[table_name][column_name].update(**changes)

    def create_index(
        self,
        table_name: str,
        index_name: str,
        fields: List[str],
        expressions: List[str],
        unique: bool,
        **kw: Any
    ):
        self.tables[table_name].indexes[index_name] = MetaIndex(
            table_name, index_name, fields, expressions, unique, **kw
        )

    def drop_index(self, table_name: str, index_name: str):
        del self.tables[table_name].indexes[index_name]

    def create_foreign_key_constraint(
        self,
        table_name: str,
        constraint_name: str,
        column_names: List[str],
        foreign_table_name: str,
        foreign_keys: List[str],
        on_delete: str
    ):
        self.tables[table_name].foreign_keys[constraint_name] = MetaForeignKey(
            table_name,
            constraint_name,
            column_names,
            foreign_table_name,
            foreign_keys,
            on_delete
        )

    def drop_foreign_key_constraint(self, table_name: str, constraint_name: str):
        del self.tables[table_name].foreign_keys[constraint_name]


class Comparator:
    def __init__(self, db: Database, meta: MetaData):
        self.db = db
        self.meta = meta

    def make_ops(self) -> List[Operation]:
        self.ops: List[Operation] = []
        self.tables()
        return self.ops

    def _build_metatable(self, dbtable: Table):
        return MetaTable(
            dbtable._tablename,
            [
                Column.from_field(field) for field in list(dbtable)
            ],
            primary_keys=list(dbtable._primary_keys)
        )

    def _build_metaindex(self, dbtable: Table, index_name: str) -> MetaIndex:
        model = dbtable._model_
        dbindex = model._indexes_[index_name]
        kw = {}
        with self.db._adapter.index_expander():
            if 'where' in dbindex:
                kw['where'] = str(dbindex['where'])
            rv = MetaIndex(
                model.tablename,
                index_name,
                [field for field in dbindex['fields']],
                [str(expr) for expr in dbindex['expressions']],
                dbindex['unique'],
                **kw
            )
        return rv

    def _build_metafk(self, dbtable: Table, fk_name: str) -> MetaForeignKey:
        model = dbtable._model_
        dbfk = model._foreign_keys_[fk_name]
        return MetaForeignKey(
            model.tablename,
            fk_name,
            dbfk['fields_local'],
            dbfk['table'],
            dbfk['fields_foreign'],
            dbfk['on_delete']
        )

    def tables(self):
        db_table_names = OrderedSet([t._tablename for t in self.db])
        meta_table_names = OrderedSet(list(self.meta.tables))
        #: new tables
        for table_name in db_table_names.difference(meta_table_names):
            meta_table = self._build_metatable(self.db[table_name])
            self.ops.append(CreateTableOp.from_table(meta_table))
            self.indexes_and_uniques(self.db[table_name], meta_table)
            self.foreign_keys(self.db[table_name], meta_table)
        #: removed tables
        for table_name in meta_table_names.difference(db_table_names):
            #: remove table indexes too
            metatable = self.meta.tables[table_name]
            for idx in metatable.indexes.values():
                self.ops.append(DropIndexOp.from_index(idx))
            #: remove table
            self.ops.append(DropTableOp.from_table(self.meta.tables[table_name]))
        #: existing tables
        for table_name in meta_table_names.intersection(db_table_names):
            self.columns(self.db[table_name], self.meta.tables[table_name])
            self.table(self.db[table_name], self.meta.tables[table_name])

    def table(self, dbtable: Table, metatable: MetaTable):
        self.indexes_and_uniques(dbtable, metatable)
        self.foreign_keys(dbtable, metatable)

    def indexes_and_uniques(
        self,
        dbtable: Table,
        metatable: MetaTable,
        ops_stack: Optional[List[Operation]] = None
    ):
        ops = ops_stack if ops_stack is not None else self.ops
        db_index_names = OrderedSet(
            [idxname for idxname in dbtable._model_._indexes_.keys()]
        )
        meta_index_names = OrderedSet(list(metatable.indexes))
        #: removed indexes
        for index_name in meta_index_names.difference(db_index_names):
            ops.append(DropIndexOp.from_index(metatable.indexes[index_name]))
        #: new indexs
        for index_name in db_index_names.difference(meta_index_names):
            ops.append(
                CreateIndexOp.from_index(
                    self._build_metaindex(dbtable, index_name)
                )
            )
        #: existing indexes
        for index_name in meta_index_names.intersection(db_index_names):
            metaindex = metatable.indexes[index_name]
            dbindex = self._build_metaindex(dbtable, index_name)
            if any(
                getattr(metaindex, key) != getattr(dbindex, key)
                for key in ['fields', 'expressions', 'unique', 'kw']
            ):
                ops.append(DropIndexOp.from_index(metaindex))
                ops.append(CreateIndexOp.from_index(dbindex))
        # TODO: uniques

    def foreign_keys(
        self,
        dbtable: Table,
        metatable: MetaTable,
        ops_stack: Optional[List[Operation]] = None
    ):
        ops = ops_stack if ops_stack is not None else self.ops
        db_fk_names = OrderedSet(
            [fkname for fkname in dbtable._model_._foreign_keys_.keys()]
        )
        meta_fk_names = OrderedSet(list(metatable.foreign_keys))
        #: removed fks
        for fk_name in meta_fk_names.difference(db_fk_names):
            ops.append(
                DropForeignKeyConstraintOp.from_foreign_key(
                    metatable.foreign_keys[fk_name]
                )
            )
        #: new fks
        for fk_name in db_fk_names.difference(meta_fk_names):
            ops.append(
                CreateForeignKeyConstraintOp.from_foreign_key(
                    self._build_metafk(dbtable, fk_name)
                )
            )
        #: existing fks
        for fk_name in meta_fk_names.intersection(db_fk_names):
            metafk = metatable.foreign_keys[fk_name]
            dbfk = self._build_metafk(dbtable, fk_name)
            if metafk != dbfk:
                ops.append(DropForeignKeyConstraintOp.from_foreign_key(metafk))
                ops.append(CreateForeignKeyConstraintOp.from_foreign_key(dbfk))

    def columns(self, dbtable: Table, metatable: MetaTable):
        db_column_names = OrderedSet([fname for fname in dbtable.fields])
        meta_column_names = OrderedSet(metatable.fields)
        #: new columns
        for column_name in db_column_names.difference(meta_column_names):
            self.ops.append(AddColumnOp.from_column_and_tablename(
                dbtable._tablename, Column.from_field(dbtable[column_name])
            ))
        #: existing columns
        for column_name in meta_column_names.intersection(db_column_names):
            self.ops.append(AlterColumnOp(dbtable._tablename, column_name))
            self.column(
                Column.from_field(dbtable[column_name]),
                metatable.columns[column_name]
            )
            if not self.ops[-1].has_changes():
                self.ops.pop()
        #: removed columns
        for column_name in meta_column_names.difference(db_column_names):
            self.ops.append(
                DropColumnOp.from_column_and_tablename(
                    dbtable._tablename, metatable.columns[column_name]
                )
            )

    def column(self, dbcolumn: Column, metacolumn: Column):
        self.notnulls(dbcolumn, metacolumn)
        self.types(dbcolumn, metacolumn)
        self.lengths(dbcolumn, metacolumn)
        self.defaults(dbcolumn, metacolumn)

    def types(self, dbcolumn: Column, metacolumn: Column):
        self.ops[-1].existing_type = metacolumn.type
        if dbcolumn.type != metacolumn.type:
            self.ops[-1].modify_type = dbcolumn.type
        if dbcolumn.geometry_type and metacolumn.geometry_type:
            for key in ("geometry_type", "srid", "dimension"):
                self.ops[-1].kw[f"existing_{key}"] = metacolumn[key]
                if dbcolumn[key] != metacolumn[key]:
                    self.ops[-1].kw[f"modify_{key}"] = dbcolumn[key]

    def lengths(self, dbcolumn: Column, metacolumn: Column):
        self.ops[-1].existing_length = metacolumn.length
        if any(
            field.type == "string" for field in [dbcolumn, metacolumn]
        ) and dbcolumn.length != metacolumn.length:
            self.ops[-1].modify_length = dbcolumn.length

    def notnulls(self, dbcolumn: Column, metacolumn: Column):
        self.ops[-1].existing_notnull = metacolumn.notnull
        if dbcolumn.notnull != metacolumn.notnull:
            self.ops[-1].modify_notnull = dbcolumn.notnull

    def defaults(self, dbcolumn: Column, metacolumn: Column):
        self.ops[-1].existing_default = metacolumn.default
        if dbcolumn.default != metacolumn.default:
            self.ops[-1].modify_default = dbcolumn.default

    @classmethod
    def compare(cls, db: Database, meta: MetaData) -> UpgradeOps:
        ops = cls(db, meta).make_ops()
        return UpgradeOps(ops)


class Generator:
    def __init__(self, db: Database, scriptdir: ScriptDir, head: str):
        self.db = db
        self.scriptdir = scriptdir
        self.head = head
        self.meta = MetaData()
        self._load_head_to_meta()

    def _load_head_to_meta(self):
        for revision in reversed(
            list(self.scriptdir.walk_revisions("base", self.head))
        ):
            migration = revision.migration_class(
                None, self.meta, is_meta=True
            )
            if migration.skip_on_compare:
                continue
            migration.up()

    def generate(self) -> UpgradeOps:
        return Comparator.compare(self.db, self.meta)

    @classmethod
    def generate_from(
        cls,
        dal: Database,
        scriptdir: ScriptDir,
        head: str
    ) -> UpgradeOps:
        return cls(dal, scriptdir, head).generate()


class Renderer:
    def render_op(self, op: Operation) -> str:
        op_renderer = renderers.dispatch(op)
        return op_renderer(op)

    def render_opcontainer(self, op_container: OpContainer) -> List[str]:
        rv = []
        if not op_container.ops:
            rv.append("pass")
        else:
            for op in op_container.ops:
                rv.append(self.render_op(op))
        return rv

    @classmethod
    def render_migration(cls, migration_op: MigrationOp):
        r = cls()
        return (
            r.render_opcontainer(migration_op.upgrade_ops),
            r.render_opcontainer(migration_op.downgrade_ops)
        )


renderers = Dispatcher()


@renderers.dispatch_for(CreateTableOp)
def _add_table(op: CreateTableOp) -> str:
    table = op.to_table()

    args = [
        col for col in [_render_column(col) for col in table.columns.values()]
        if col
    ]
    # + sorted([
    #     rcons for rcons in [
    #         _render_constraint(cons) for cons in table.constraints]
    #         if rcons is not None
    #     ]
    # )
    indent = " " * 12

    if len(args) > 255:
        args = '*[' + (',\n' + indent).join(args) + ']'
    else:
        args = (',\n' + indent).join(args)

    text = (
        "self.create_table(\n" + indent + "%(tablename)r,\n" + indent + "%(args)s,\n" +
        indent + "primary_keys=%(primary_keys)r"
    ) % {
        'tablename': op.table_name,
        'args': args,
        'primary_keys': table.primary_keys
    }
    for k in sorted(op.kw):
        text += ",\n" + indent + "%s=%r" % (k.replace(" ", "_"), op.kw[k])
    text += ")"
    return text


@renderers.dispatch_for(DropTableOp)
def _drop_table(op: DropTableOp) -> str:
    text = "self.drop_table(%(tname)r" % {
        "tname": op.table_name
    }
    text += ")"
    return text


def _render_column(column: Column) -> str:
    opts = []

    if column.default is not None:
        opts.append(("default", column.default))

    # if not column.autoincrement:
    #     opts.append(("autoincrement", column.autoincrement))

    if column.notnull:
        opts.append(("notnull", column.notnull))

    if column.type in ("string", "password", "upload"):
        opts.append(("length", column.length))
    elif column.type.startswith('reference'):
        opts.append(("ondelete", column.ondelete))
    elif column.type.startswith("geo"):
        for key in ("geometry_type", "srid", "dimension"):
            if column[key] is not None:
                opts.append((key, column[key]))

    kw_str = ""
    if opts:
        kw_str = ", %s" % \
            ", ".join(["%s=%r" % (key, val) for key, val in opts])
    return "migrations.Column(%(name)r, %(type)r%(kw)s)" % {
        'name': column.name,
        'type': column.type,
        'kw': kw_str
    }


@renderers.dispatch_for(AddColumnOp)
def _add_column(op: AddColumnOp) -> str:
    return "self.add_column(%(tname)r, %(column)s)" % {
        "tname": op.table_name,
        "column": _render_column(op.column)
    }


@renderers.dispatch_for(DropColumnOp)
def _drop_column(op: DropTableOp) -> str:
    return "self.drop_column(%(tname)r, %(cname)r)" % {
        "tname": op.table_name,
        "cname": op.column_name
    }


@renderers.dispatch_for(AlterColumnOp)
def _alter_column(op: AlterColumnOp) -> str:
    indent = " " * 12
    text = "self.alter_column(%(tname)r, %(cname)r" % {
        'tname': op.table_name,
        'cname': op.column_name}

    if op.existing_type is not None:
        text += ",\n%sexisting_type=%r" % (indent, op.existing_type)
    if op.existing_length:
        text += ",\n%sexisting_length=%r" % (indent, op.existing_length)
    if op.modify_default is not DEFAULT_VALUE:
        text += ",\n%sdefault=%r" % (indent, op.modify_default)
    if op.modify_type is not None:
        text += ",\n%stype=%r" % (indent, op.modify_type)
    if op.modify_length is not None:
        text += ",\n%slength=%r" % (indent, op.modify_length)
    if op.modify_notnull is not None:
        text += ",\n%snotnull=%r" % (indent, op.modify_notnull)
    if op.modify_notnull is None and op.existing_notnull is not None:
        text += ",\n%sexisting_notnull=%r" % (indent, op.existing_notnull)
    if op.modify_default is DEFAULT_VALUE and op.existing_default:
        text += ",\n%sexisting_default=%s" % (indent, op.existing_default)
    for key, val in op.kw.items():
        if key.startswith("existing_") or key.startswith("modify_"):
            text += ",\n%s%s=%r" % (indent, key, val)

    text += ")"
    return text


@renderers.dispatch_for(CreateIndexOp)
def _add_index(op: CreateIndexOp) -> str:
    kw_str = ""
    if op.kw:
        kw_str = ", %s" % ", ".join(
            ["%s=%r" % (key, val) for key, val in op.kw.items()])
    return "self.create_index(%(iname)r, %(tname)r, %(idata)s)" % {
        "tname": op.table_name,
        "iname": op.index_name,
        "idata": "%r, expressions=%r, unique=%s%s" % (
            op.fields, op.expressions, op.unique, kw_str)
    }


@renderers.dispatch_for(DropIndexOp)
def _drop_index(op: DropIndexOp) -> str:
    return "self.drop_index(%(iname)r, %(tname)r)" % {
        "tname": op.table_name,
        "iname": op.index_name
    }


@renderers.dispatch_for(CreateForeignKeyConstraintOp)
def _add_fk_constraint(op: CreateForeignKeyConstraintOp) -> str:
    kw_str = ""
    if op.kw:
        kw_str = ", %s" % ", ".join(
            ["%s=%r" % (key, val) for key, val in op.kw.items()]
        )
    return "self.create_foreign_key(%s%s)" % (
        "%r, %r, %r, %r, %r, on_delete=%r" % (
            op.constraint_name,
            op.table_name,
            op.foreign_table_name,
            op.column_names,
            op.foreign_keys,
            op.on_delete
        ),
        kw_str
    )


@renderers.dispatch_for(DropForeignKeyConstraintOp)
def _drop_fk_constraint(op: DropForeignKeyConstraintOp) -> str:
    return "self.drop_foreign_key(%(cname)r, %(tname)r)" % {
        "tname": op.table_name,
        "cname": op.constraint_name
    }



# -*- coding: utf-8 -*-
"""
    emmett.rsgi.helpers
    -------------------

    Provides RSGI helpers

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from granian.rsgi import WebsocketProtocol


class WSTransport:
    __slots__ = [
        'protocol', 'transport',
        'accepted', 'interrupted',
        'input', 'status', 'noop'
    ]

    def __init__(
        self,
        protocol: WebsocketProtocol
    ) -> None:
        self.protocol = protocol
        self.transport = None
        self.accepted = asyncio.Event()
        self.input = asyncio.Queue()
        self.interrupted = False
        self.status = 200
        self.noop = asyncio.Event()

    async def init(self):
        self.transport = await self.protocol.accept()
        self.accepted.set()

    @property
    def receive(self):
        return self.input.get



# -*- coding: utf-8 -*-
"""
    tests.orm_row
    -------------

    Test ORM row objects
"""

import pickle
import pytest

from uuid import uuid4

from emmett import App, sdict, now
from emmett.orm import (
    Database, Model, Field, belongs_to, has_many, refers_to, rowattr, rowmethod
)
from emmett.orm.errors import ValidationError
from emmett.orm.helpers import RowReferenceMixin
from emmett.orm.migrations.utils import generate_runtime_migration
from emmett.orm.objects import Row


class One(Model):
    has_many("twos", "threes")

    foo = Field.string(notnull=True)
    bar = Field.string()

    @rowattr("twos_list")
    def _row_attr_test(self, row):
        return [el.foo for el in row.twos()]


class Two(Model):
    belongs_to("one")

    foo = Field.string()
    bar = Field.string()


class Three(Model):
    refers_to("one")

    foo = Field.string()


class Override(Model):
    foo = Field.string()
    deleted_at = Field.datetime()

    validation = {"deleted_at": {"allow": "empty"}}

    @rowattr("test")
    def _row_attr_test(self, row):
        return "test"

    @rowmethod("destroy")
    def _row_destroy(self, row):
        row.deleted_at = now()
        row.save()

    @rowmethod("force_destroy")
    def _row_force_destroy(self, row):
        self.super_rowmethod("destroy")(row)


class Crypted(Model):
    foo = Field.string()
    bar = Field.password()


@pytest.fixture(scope='module')
def _db():
    app = App(__name__)
    db = Database(
        app,
        config=sdict(
            uri=f'sqlite://{uuid4().hex}.db',
            auto_connect=True
        )
    )
    db.define_models(One, Two, Three, Override, Crypted)
    return db

@pytest.fixture(scope='function')
def db(_db):
    migration = generate_runtime_migration(_db)
    migration.up()
    yield _db
    migration.down()


def test_rowclass(db):
    ret = db.One.insert(foo="test1", bar="test2")
    db.Two.insert(one=ret, foo="test1", bar="test2")

    ret._allocate_()
    assert type(ret._refrecord) == One._instance_()._rowclass_

    row = One.get(ret.id)
    assert type(row) == One._instance_()._rowclass_

    row = One.first()
    assert type(row) == One._instance_()._rowclass_

    row = db(db.One).select().first()
    assert type(row) == One._instance_()._rowclass_

    row = db(db.One).select(db.One.ALL).first()
    assert type(row) == One._instance_()._rowclass_

    row = One.all().select().first()
    assert type(row) == One._instance_()._rowclass_

    row = One.where(lambda m: m.id != None).select().first()
    assert type(row) == One._instance_()._rowclass_

    row = db(db.One).select().first()
    assert type(row) == One._instance_()._rowclass_

    row = db(db.One).select(db.One.ALL).first()
    assert type(row) == One._instance_()._rowclass_

    row = One.all().select(One.bar).first()
    assert type(row) == Row

    row = db(db.One).select(One.bar).first()
    assert type(row) == Row

    row = One.all().join("twos").select().first()
    assert type(row) == One._instance_()._rowclass_
    assert type(row.twos().first()) == Two._instance_()._rowclass_

    row = One.all().join("twos").select(One.table.ALL, Two.table.ALL).first()
    assert type(row) == One._instance_()._rowclass_
    assert type(row.twos().first()) == Two._instance_()._rowclass_

    row = One.all().join("twos").select(One.foo, Two.foo).first()
    assert type(row) == Row
    assert type(row.ones) == Row
    assert type(row.twos) == Row

    row = db(Two.one == One.id).select().first()
    assert type(row) == Row
    assert type(row.ones) == One._instance_()._rowclass_
    assert type(row.twos) == Two._instance_()._rowclass_

    row = db(Two.one == One.id).select(One.table.ALL, Two.foo).first()
    assert type(row) == Row
    assert type(row.ones) == One._instance_()._rowclass_
    assert type(row.twos) == Row

    row = db(Two.one == One.id).select(One.foo, Two.foo).first()
    assert type(row) == Row
    assert type(row.ones) == Row
    assert type(row.twos) == Row

    for row in db(Two.one == One.id).iterselect():
        assert type(row) == Row
        assert type(row.ones) == One._instance_()._rowclass_
        assert type(row.twos) == Two._instance_()._rowclass_

    for row in db(Two.one == One.id).iterselect(One.table.ALL, Two.foo):
        assert type(row) == Row
        assert type(row.ones) == One._instance_()._rowclass_
        assert type(row.twos) == Row

    for row in db(Two.one == One.id).iterselect(One.foo, Two.foo):
        assert type(row) == Row
        assert type(row.ones) == Row
        assert type(row.twos) == Row


def test_concrete(db):
    row = One.new(foo="test")
    assert not row._concrete

    row.save()
    assert row._concrete

    row = One.get(row.id)
    assert row._concrete


def test_changes(db):
    row = One.new(foo="test1")
    assert not row.has_changed

    row.bar = "test2"
    assert row.has_changed
    assert row.has_changed_value("bar")
    assert row.get_value_change("bar") == (None, "test2")

    row.bar = "test2a"
    assert row.has_changed
    assert row.has_changed_value("bar")
    assert row.get_value_change("bar") == (None, "test2a")

    row.bar = None
    assert not row.has_changed
    assert not row.has_changed_value("bar")
    assert row.get_value_change("bar") is None

    row.update(bar="test2b")
    assert row.has_changed
    assert row.has_changed_value("bar")
    assert row.get_value_change("bar") == (None, "test2b")

    row.update({"bar": "test2c"})
    assert row.has_changed
    assert row.has_changed_value("bar")
    assert row.get_value_change("bar") == (None, "test2c")

    row.update(bar=None)
    assert not row.has_changed
    assert not row.has_changed_value("bar")
    assert row.get_value_change("bar") is None

    row.bar = "test2"
    row.save()
    assert not row.has_changed


def test_validation_methods(db):
    row = One.new()
    assert not row.is_valid
    assert set(row.validation_errors.keys()).issubset({"foo"})
    assert not row.save()
    with pytest.raises(ValidationError):
        row.save(raise_on_error=True)

    row.foo = "test"
    assert row.is_valid
    assert not row.validation_errors
    assert row.save()


def test_clone_methods(db):
    row = One.new(foo="test1")
    row.bar = "test2"

    row2 = row.clone()
    row3 = row.clone_changed()

    assert not row2._concrete
    assert not row2.has_changed
    assert not row2.bar
    assert not row3._concrete
    assert not row3.has_changed
    assert row3.bar == "test2"

    row.save()
    row.foo = "test1a"
    row2 = row.clone()
    row3 = row.clone_changed()

    assert row2._concrete
    assert not row2.has_changed
    assert row2.foo == "test1"
    assert row3._concrete
    assert not row3.has_changed
    assert row3.foo == "test1a"


def test_refresh(db):
    row = One.new(foo="test1")
    assert not row.refresh()

    row.save()
    assert row.refresh()

    row.foo = "test2"
    assert row.refresh()
    assert row.foo == "test1"


def test_methods_override(db):
    row = Override.new(foo="test")
    row.save()
    assert row.id
    assert not row.deleted_at

    row.destroy()
    assert row.id
    assert row.deleted_at

    row.force_destroy()
    assert not row.id


def test_attributes(db):
    row = One.new(foo="test", baz="test")
    assert row._fields == {"id": None, "foo": "test", "bar": None}
    assert row.__dict__ == {"baz": "test"}

    assert row.get("foo") == "test"
    assert row.get("baz") == "test"
    assert row.get("test", "foobar") == "foobar"

    assert "foo" in row
    assert "baz" in row
    assert "test" not in row

    row.save()
    assert row.__dict__ == {"baz": "test"}

    row.foo = "test1"
    row.baz = "test1"
    assert set(row._changes.keys()) == {"foo", "baz"}

    row.save()
    assert set(row._changes.keys()) == {"baz"}

    row.refresh()
    assert not row.__dict__
    assert not row._changes


def test_relation_wrappers(db):
    r1 = db.One.insert(foo="test1")
    r2 = db.One.insert(foo="test2")

    r = Two.new()
    assert not r.one

    r = Two.new(one=None)
    assert not r.one

    r = Two.new(one=r1)
    assert isinstance(r.one, RowReferenceMixin)

    r = Two.new(one=r1.id)
    assert isinstance(r.one, RowReferenceMixin)

    r = Two.new(one=One.get(r1))
    assert isinstance(r.one, RowReferenceMixin)

    r.one = r2.id
    assert isinstance(r.one, RowReferenceMixin)

    r = Three.new()
    assert not r.one

    r = Three.new(one=None)
    assert not r.one

    r = Three.new(one=r1)
    assert isinstance(r.one, RowReferenceMixin)

    r = Three.new(one=r1.id)
    assert isinstance(r.one, RowReferenceMixin)

    r = Three.new(one=One.get(r1))
    assert isinstance(r.one, RowReferenceMixin)

    r.one = r2.id
    assert isinstance(r.one, RowReferenceMixin)


def test_password_wrapper(db):
    db.Crypted.validate_and_insert(foo="foobar", bar="foobar")

    row = Crypted.first()
    crypted_value = str(row.bar)

    row.save()
    assert str(row.bar) == crypted_value

    row.foo = "test"
    row.save()
    assert str(row.bar) == crypted_value

    row.bar = "testtest"
    row.save()
    assert str(row.bar) != crypted_value


def test_pickle(db):
    r1 = One.new(foo="test")
    r1.save()
    r2 = Two.new(one=r1, foo="test")
    r2.save()
    r3 = Override.new(foo="test")
    r3.save()

    assert r1.twos_list == ["test"]
    assert r2.one.foo == "test"
    assert r3.test == "test"

    r1.inject = "test"
    r2.inject = "test"
    r3.inject = "test"

    d1 = pickle.dumps(r1)
    l1 = pickle.loads(d1)

    assert l1._fields == r1._fields
    assert l1.__dict__ == r1.__dict__

    d2 = pickle.dumps(r2)
    l2 = pickle.loads(d2)

    assert l2._fields == r2._fields
    assert l2.__dict__ == r2.__dict__

    d3 = pickle.dumps(r3)
    l3 = pickle.loads(d3)

    assert l3._fields == r3._fields
    assert l3.__dict__ == r3.__dict__



# -*- coding: utf-8 -*-
"""
    emmett.asgi.helpers
    -------------------

    Provides ASGI helpers

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""


class RequestCancelled(Exception):
    ...



# -*- coding: utf-8 -*-
"""
    tests.orm_gis
    -------------

    Test ORM GIS features
"""

import os
import pytest

from emmett import App, sdict
from emmett.orm import Database, Model, Field, geo
from emmett.orm.migrations.utils import generate_runtime_migration

require_postgres = pytest.mark.skipif(
    not os.environ.get("POSTGRES_URI"), reason="No postgres database"
)


class Geography(Model):
    name = Field.string()
    geo = Field.geography()
    point = Field.geography("POINT")
    line = Field.geography("LINESTRING")
    polygon = Field.geography("POLYGON")
    multipoint = Field.geography("MULTIPOINT")
    multiline = Field.geography("MULTILINESTRING")
    multipolygon = Field.geography("MULTIPOLYGON")


class Geometry(Model):
    name = Field.string()
    geo = Field.geometry()
    point = Field.geometry("POINT")
    line = Field.geometry("LINESTRING")
    polygon = Field.geometry("POLYGON")
    multipoint = Field.geometry("MULTIPOINT")
    multiline = Field.geometry("MULTILINESTRING")
    multipolygon = Field.geometry("MULTIPOLYGON")


@pytest.fixture(scope='module')
def _db():
    app = App(__name__)
    db = Database(
        app,
        config=sdict(
            uri=f"postgres://{os.environ.get('POSTGRES_URI')}"
        )
    )
    db.define_models(
        Geography,
        Geometry
    )
    return db


@pytest.fixture(scope='function')
def db(_db):
    migration = generate_runtime_migration(_db)
    with _db.connection():
        migration.up()
        yield _db
        migration.down()


@require_postgres
def test_field_types(_db):
    assert Geography.geo.type == "geography(GEOMETRY,4326,2)"
    assert Geography.point.type == "geography(POINT,4326,2)"
    assert Geography.line.type == "geography(LINESTRING,4326,2)"
    assert Geography.polygon.type == "geography(POLYGON,4326,2)"
    assert Geography.multipoint.type == "geography(MULTIPOINT,4326,2)"
    assert Geography.multiline.type == "geography(MULTILINESTRING,4326,2)"
    assert Geography.multipolygon.type == "geography(MULTIPOLYGON,4326,2)"
    assert Geometry.geo.type == "geometry(GEOMETRY,4326,2)"
    assert Geometry.point.type == "geometry(POINT,4326,2)"
    assert Geometry.line.type == "geometry(LINESTRING,4326,2)"
    assert Geometry.polygon.type == "geometry(POLYGON,4326,2)"
    assert Geometry.multipoint.type == "geometry(MULTIPOINT,4326,2)"
    assert Geometry.multiline.type == "geometry(MULTILINESTRING,4326,2)"
    assert Geometry.multipolygon.type == "geometry(MULTIPOLYGON,4326,2)"


@require_postgres
def test_gis_insert(db):
    for model in [Geometry, Geography]:
        row = model.new(
            point=geo.Point(1, 1),
            line=geo.Line((0, 0), (20, 80), (80, 80)),
            polygon=geo.Polygon((0, 0), (150, 0), (150, 10), (0, 10), (0, 0)),
            multipoint=geo.MultiPoint((1, 1), (2, 2)),
            multiline=geo.MultiLine(((1, 1), (2, 2), (3, 3)), ((1, 1), (4, 4), (5, 5))),
            multipolygon=geo.MultiPolygon(
                (
                    ((0, 0), (20, 0), (20, 20), (0, 0)),
                    ((0, 0), (30, 0), (30, 30), (0, 0))
                ),
                (
                    ((1, 1), (21, 1), (21, 21), (1, 1)),
                    ((1, 1), (31, 1), (31, 31), (1, 1))
                )
            )
        )
        row.save()

        assert row.point == "POINT(1.000000 1.000000)"
        assert row.point.geometry == "POINT"
        assert row.point.coordinates == (1, 1)
        assert not row.point.groups

        assert row.line == "LINESTRING({})".format(
            ",".join([
                " ".join(f"{v}.000000" for v in tup)
                for tup in [
                    (0, 0),
                    (20, 80),
                    (80, 80)
                ]
            ])
        )
        assert row.line.geometry == "LINESTRING"
        assert row.line.coordinates == ((0, 0), (20, 80), (80, 80))
        assert not row.line.groups

        assert row.polygon == "POLYGON(({}))".format(
            ",".join([
                " ".join(f"{v}.000000" for v in tup)
                for tup in [
                    (0, 0),
                    (150, 0),
                    (150, 10),
                    (0, 10),
                    (0, 0)
                ]
            ])
        )
        assert row.polygon.geometry == "POLYGON"
        assert row.polygon.coordinates == (
            ((0, 0), (150, 0), (150, 10), (0, 10), (0, 0)),
        )
        assert not row.polygon.groups

        assert row.multipoint == "MULTIPOINT((1.000000 1.000000),(2.000000 2.000000))"
        assert row.multipoint.geometry == "MULTIPOINT"
        assert row.multipoint.coordinates == ((1, 1), (2, 2))
        assert len(row.multipoint.groups) == 2
        assert row.multipoint.groups[0] == geo.Point(1, 1)
        assert row.multipoint.groups[1] == geo.Point(2, 2)

        assert row.multiline == "MULTILINESTRING({})".format(
            ",".join([
                "({})".format(
                    ",".join([
                        " ".join(f"{v}.000000" for v in tup)
                        for tup in group
                    ])
                ) for group in [
                    ((1, 1), (2, 2), (3, 3)),
                    ((1, 1), (4, 4), (5, 5))
                ]
            ])
        )
        assert row.multiline.geometry == "MULTILINESTRING"
        assert row.multiline.coordinates == (
            ((1, 1), (2, 2), (3, 3)),
            ((1, 1), (4, 4), (5, 5))
        )
        assert len(row.multiline.groups) == 2
        assert row.multiline.groups[0] == geo.Line((1, 1), (2, 2), (3, 3))
        assert row.multiline.groups[1] == geo.Line((1, 1), (4, 4), (5, 5))

        assert row.multipolygon == "MULTIPOLYGON({})".format(
            ",".join([
                "({})".format(
                    ",".join([
                        "({})".format(
                            ",".join([
                                " ".join(f"{v}.000000" for v in tup)
                                for tup in group
                            ])
                        ) for group in polygon
                    ])
                ) for polygon in [
                    (
                        ((0, 0), (20, 0), (20, 20), (0, 0)),
                        ((0, 0), (30, 0), (30, 30), (0, 0))
                    ),
                    (
                        ((1, 1), (21, 1), (21, 21), (1, 1)),
                        ((1, 1), (31, 1), (31, 31), (1, 1))
                    )
                ]
            ])
        )
        assert row.multipolygon.geometry == "MULTIPOLYGON"
        assert row.multipolygon.coordinates == (
            (
                ((0, 0), (20, 0), (20, 20), (0, 0)),
                ((0, 0), (30, 0), (30, 30), (0, 0))
            ),
            (
                ((1, 1), (21, 1), (21, 21), (1, 1)),
                ((1, 1), (31, 1), (31, 31), (1, 1))
            )
        )
        assert len(row.multipolygon.groups) == 2
        assert row.multipolygon.groups[0] == geo.Polygon(
            ((0, 0), (20, 0), (20, 20), (0, 0)),
            ((0, 0), (30, 0), (30, 30), (0, 0))
        )
        assert row.multipolygon.groups[1] == geo.Polygon(
            ((1, 1), (21, 1), (21, 21), (1, 1)),
            ((1, 1), (31, 1), (31, 31), (1, 1))
        )


@require_postgres
def test_gis_select(db):
    for model in [Geometry, Geography]:
        row = model.new(
            point=geo.Point(1, 1),
            line=geo.Line((0, 0), (20, 80), (80, 80)),
            polygon=geo.Polygon((0, 0), (150, 0), (150, 10), (0, 10), (0, 0)),
            multipoint=geo.MultiPoint((1, 1), (2, 2)),
            multiline=geo.MultiLine(((1, 1), (2, 2), (3, 3)), ((1, 1), (4, 4), (5, 5))),
            multipolygon=geo.MultiPolygon(
                (
                    ((0, 0), (20, 0), (20, 20), (0, 0)),
                    ((0, 0), (30, 0), (30, 30), (0, 0))
                ),
                (
                    ((1, 1), (21, 1), (21, 21), (1, 1)),
                    ((1, 1), (31, 1), (31, 31), (1, 1))
                )
            )
        )
        row.save()
        row = model.get(row.id)

        assert row.point.geometry == "POINT"
        assert row.point.coordinates == (1, 1)
        assert not row.point.groups

        assert row.line.geometry == "LINESTRING"
        assert row.line.coordinates == ((0, 0), (20, 80), (80, 80))
        assert not row.line.groups

        assert row.polygon.geometry == "POLYGON"
        assert row.polygon.coordinates == (
            ((0, 0), (150, 0), (150, 10), (0, 10), (0, 0)),
        )
        assert not row.polygon.groups

        assert row.multipoint.geometry == "MULTIPOINT"
        assert row.multipoint.coordinates == ((1, 1), (2, 2))
        assert len(row.multipoint.groups) == 2
        assert row.multipoint.groups[0] == geo.Point(1, 1)
        assert row.multipoint.groups[1] == geo.Point(2, 2)

        assert row.multiline.geometry == "MULTILINESTRING"
        assert row.multiline.coordinates == (
            ((1, 1), (2, 2), (3, 3)),
            ((1, 1), (4, 4), (5, 5))
        )
        assert len(row.multiline.groups) == 2
        assert row.multiline.groups[0] == geo.Line((1, 1), (2, 2), (3, 3))
        assert row.multiline.groups[1] == geo.Line((1, 1), (4, 4), (5, 5))

        assert row.multipolygon.geometry == "MULTIPOLYGON"
        assert row.multipolygon.coordinates == (
            (
                ((0, 0), (20, 0), (20, 20), (0, 0)),
                ((0, 0), (30, 0), (30, 30), (0, 0))
            ),
            (
                ((1, 1), (21, 1), (21, 21), (1, 1)),
                ((1, 1), (31, 1), (31, 31), (1, 1))
            )
        )
        assert len(row.multipolygon.groups) == 2
        assert row.multipolygon.groups[0] == geo.Polygon(
            ((0, 0), (20, 0), (20, 20), (0, 0)),
            ((0, 0), (30, 0), (30, 30), (0, 0))
        )
        assert row.multipolygon.groups[1] == geo.Polygon(
            ((1, 1), (21, 1), (21, 21), (1, 1)),
            ((1, 1), (31, 1), (31, 31), (1, 1))
        )
